<?php
/*	pb_mainwp_backupbuddy_backup class
 *	
 *	Handles the actual backup procedures.
 *	
 *	USED BY:
 *
 *	1) Full & DB backups
 *	2) Multisite backups & exports
 *
 */
class pb_mainwp_backupbuddy_backup {
	
	private $_errors = array();								// TODO:  No longer used? Remove?
	private $_status_logging_started = false;				// Marked true once anything has been status logged during this process. Used by status().
	private $_currentStepIndex = '';						// When running a backup, this is the current index in the steps array within the fileoptions.
	
	// Constants for Zip Build Strategy - here for now but will be moved to central file
	const ZIP_BUILD_STRATEGY_SBSS = 2;						// Single-Burst/Single-Step
	const ZIP_BUILD_STRATEGY_MBSS = 3;						// Muti-Burst/Single-Step
	const ZIP_BUILD_STRATEGY_MBMS = 4;						// Multi-Burst/Multi-Step
	const ZIP_BUILD_STRATEGY_MIN = self::ZIP_BUILD_STRATEGY_SBSS;
	const ZIP_BUILD_STRATEGY_MAX = self::ZIP_BUILD_STRATEGY_MBMS;
	
	
	
	/*	__construct()
	 *	
	 *	Default contructor. Initialized core and zipbuddy classes.
	 *	
	 *	@return		null
	 */
	function __construct() {
		
		// Load core if it has not been instantiated yet.
		if ( ! class_exists( 'mainwp_backupbuddy_core' ) ) {
			require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/core.php' );
		}
		
		
		// Register PHP shutdown function to help catch and log fatal PHP errors during backup.
		register_shutdown_function( array( &$this, 'shutdown_function' ) );
		
	} // End __construct().
	
	
	
	/*	shutdown_function()
	 *	
	 *	Used for catching fatal PHP errors during backup to write to log for debugging.
	 *	
	 *	@return		null
	 */
	public function shutdown_function() {
		
		
		// Get error message.
		// Error types: http://php.net/manual/en/errorfunc.constants.php
		$e = error_get_last();
		if ( $e === NULL ) { // No error of any kind.
			return;
		} else { // Some type of error.
			if ( !is_array( $e ) || ( $e['type'] != E_ERROR ) && ( $e['type'] != E_USER_ERROR ) ) { // Return if not a fatal error.
				return;
			}
		}
		
		
		// Calculate log directory.
		$log_directory = mainwp_backupbuddy_core::getLogDirectory();
		$main_file = $log_directory . 'log-' . pb_mainwp_backupbuddy::$options['log_serial'] . '.txt';
		
		
		// Determine if writing to a serial log.
		if ( pb_mainwp_backupbuddy::$_status_serial != '' ) {
			$serial = pb_mainwp_backupbuddy::$_status_serial;
			$serial_file = $log_directory . 'status-' . $serial . '_' . pb_mainwp_backupbuddy::$options['log_serial'] . '.txt';
			$write_serial = true;
		} else {
			$write_serial = false;
		}
		
		
		// Format error message.
		$e_string = 'PHP_ERROR ' . __( 'Error #32893. Fatal PHP error encountered:', 'it-l10n-mainwp-backupbuddy' );
		foreach( (array)$e as $e_line_title => $e_line ) {
			$e_string .= $e_line_title . ' => ' . $e_line . "; ";
		}
		$e_string .= ".\n";
		
		
		// Write to log.
		@file_put_contents( $main_file, $e_string, FILE_APPEND );
		if ( $write_serial === true ) {
			@file_put_contents( $serial_file, $e_string, FILE_APPEND );
		}
		
		
	} // End shutdown_function.
	
	
	
	
	
	
		
	
	/*	pre_backup()
	 *	
	 *	Set up the backup data structure containing steps, set up temp directories, etc.
	 *	
	 *	@param		array		$profile						Backup profile array data. Prev (pre-4.0):	Backup type. Valid values: db, full, export.
	 *	@param		string		$trigger						What triggered this backup. Valid values: scheduled, manual.
	 *	@param		array		$pre_backup						Array of functions to prepend to the backup steps array.
	 *	@param		array		$post_backup					Array of functions to append to the backup steps array. Ie. sending to remote destination.
	 *	@param		string		$schedule_title					Title name of schedule. Used for tracking what triggered this in logging. For debugging.
	 *	@param		array		$export_plugins					For use in export backup type. List of plugins to export.
	 *	@param		string		$deployDirection				blank for not deploy, push, or pull.
	 @	@param		array|''	$deployDestinationSettings		Destination settings for the deployment. Empty string when not deployment.
	 *	@return		boolean										True on success; false otherwise.
	 */
	function pre_backup( $serial, $archiveFile, $profile, $trigger, $pre_backup = array(), $post_backup = array(), $schedule_title = '', $export_plugins = array(), $deployDirection, $deployDestinationSettings ) {
		
		pb_mainwp_backupbuddy::status( 'startFunction', json_encode( array( 'function' => 'pre_backup', 'title' => 'Getting ready to backup' ) ) );
		
		$type = $profile['type'];
		
		// Log some status information.
		pb_mainwp_backupbuddy::status( 'details', __( 'Performing pre-backup procedures.', 'it-l10n-mainwp-backupbuddy' ) );
		if ( $type == 'full' ) {
			pb_mainwp_backupbuddy::status( 'message', __( 'Full backup mode.', 'it-l10n-mainwp-backupbuddy' ) );
		} elseif ( $type == 'db' ) {
			pb_mainwp_backupbuddy::status( 'message', __( 'Database only backup mode.', 'it-l10n-mainwp-backupbuddy' ) );
		} elseif ( $type == 'files' ) {
			pb_mainwp_backupbuddy::status( 'message', __( 'Files only backup mode.', 'it-l10n-mainwp-backupbuddy' ) );
			//$profile['skip_database_dump'] = '1';
		} elseif ( $type == 'export' ) {
			pb_mainwp_backupbuddy::status( 'message', __( 'Multisite subsite export mode.', 'it-l10n-mainwp-backupbuddy' ) );
		} else {
			pb_mainwp_backupbuddy::status( 'error', 'Error #8587383: Unknown backup mode `' . htmlentities( $type ) . '`.' );
		}
		
		if ( '' != $deployDirection ) {
			pb_mainwp_backupbuddy::status( 'details', 'Deployment direction: `' . $deployDirection . '`.' );
		}
		
		if ( '1' == pb_mainwp_backupbuddy::$options['prevent_flush'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'Flushing will be skipped based on advanced settings.' );
		} else {
			pb_mainwp_backupbuddy::status( 'details', 'Flushing will not be skipped (default).' );
		}
		
		// Schedule daily housekeeping.
		mainwp_backupbuddy_core::verifyHousekeeping();
		
		// Verify directories.
		pb_mainwp_backupbuddy::status( 'details', 'Verifying directories ...' );
		if ( false === mainwp_backupbuddy_core::verify_directories() ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #18573. Error verifying directories. See details above. Backup halted.' );
			pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
			die();
		} else {
			pb_mainwp_backupbuddy::status( 'details', 'Directories verified.' );
		}
		
		// Delete all backup archives if this troubleshooting option is enabled.
		if ( pb_mainwp_backupbuddy::$options['delete_archives_pre_backup'] == '1' ) {
			pb_mainwp_backupbuddy::status( 'message', 'Deleting all existing backups prior to backup as configured on the settings page.' );
			$file_list = glob( mainwp_backupbuddy_core::getBackupDirectory() . 'backup*.zip' );
			if ( is_array( $file_list ) && !empty( $file_list ) ) {
				foreach( $file_list as $file ) {
					if ( @unlink( $file ) === true ) {
						pb_mainwp_backupbuddy::status( 'details', 'Deleted backup archive `' . basename( $file ) . '` based on settings to delete all backups.' );
					} else {
						pb_mainwp_backupbuddy::status( 'details', 'Unable to delete backup archive `' . basename( $file ) . '` based on settings to delete all backups. Verify permissions.' );
					}
				}
			}
		}
				
		// Generate unique serial ID.
		pb_mainwp_backupbuddy::status( 'details', 'Backup serial generated: `' . $serial . '`.' );
		
		pb_mainwp_backupbuddy::status( 'details', 'About to load fileoptions data in create mode.' );
		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/fileoptions.php' );
		pb_mainwp_backupbuddy::status( 'details', 'Fileoptions instance #40.' );
		$this->_backup_options = new pb_mainwp_backupbuddy_fileoptions( mainwp_backupbuddy_core::getLogDirectory() . 'fileoptions/' . $serial . '.txt', $read_only = false, $ignore_lock = false, $create_file = true );
		if ( true !== ( $result = $this->_backup_options->is_ok() ) ) {
			pb_mainwp_backupbuddy::status( 'error', __('Fatal Error #9034 A. Unable to access fileoptions data.', 'it-l10n-mainwp-backupbuddy' ) . ' Error: ' . $result );
			pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
			return false;
		}
		pb_mainwp_backupbuddy::status( 'details', 'Fileoptions data loaded.' );
		$this->_backup = &$this->_backup_options->options; // Set reference.
		
		
		// Cleanup internal stats. Deployments should not impact stats.
		if ( '' == $deployDirection ) {
			pb_mainwp_backupbuddy::status( 'details', 'Updating statistics for last backup start.' );
			pb_mainwp_backupbuddy::$options['last_backup_start'] = microtime(true); // Reset time since last backup.
			pb_mainwp_backupbuddy::$options['last_backup_serial'] = $serial;
			pb_mainwp_backupbuddy::save();
		}
		
		
		// Output active plugins list for debugging...
		$activePlugins = get_option( 'active_plugins' );
		pb_mainwp_backupbuddy::status( 'details', 'Active WordPress plugins (' . count( $activePlugins ) . '): `' . implode( '; ', $activePlugins ) . '`.' );
		pb_mainwp_backupbuddy::status( 'startSubFunction', json_encode( array( 'function' => 'wp_plugins_found', 'title' => 'Found ' . count( $activePlugins ) . ' active WordPress plugins.' ) ) );
		unset( $activePlugins );
		
		
		// Compression to bool.
		/*
		if ( $profile['compression'] == '1' ) {
			$profile['compression'] = true;
		} else {
			$profile['compression'] = false;
		}
		*/
		if ( pb_mainwp_backupbuddy::$options['compression'] == '1' ) {
			$compression = true;
		} else {
			$compression = false;
		}
		
		
		
		$archiveURL = '';
		$abspath = str_replace( '\\', '/', ABSPATH ); // Change slashes to handle Windows as we store backup_directory with Linux-style slashes even on Windows.
		$backup_dir = str_replace( '\\', '/', mainwp_backupbuddy_core::getBackupDirectory() );
		if ( FALSE !== stristr( $backup_dir, $abspath ) ) { // Make sure file to download is in a publicly accessible location (beneath WP web root technically).
			$sitepath = str_replace( $abspath, '', $backup_dir );
			$archiveURL = rtrim( site_url(), '/\\' ) . '/' . trim( $sitepath, '/\\' ) . '/' . basename( $archiveFile );
		}
		
		$forceSingleDatabaseFile = false;
		if ( '1' == pb_mainwp_backupbuddy::$options['force_single_db_file'] ) {
			$forceSingleDatabaseFile = true;
		}
		
		// Set up the backup data.
		$this->_backup = array(
			'data_version'			=>		1,												// Data structure version. Upped to 1 for BBv5.0.
			'backupbuddy_version'	=>		pb_mainwp_backupbuddy::settings( 'version' ),			// BB version used for this backup.
			'serial'				=>		$serial,										// Unique identifier.
			'init_complete'			=>		false,											// Whether pre_backup() completed or not. Other step status is already tracked and stored in data structure but pre_backup 'step' was not until now. Jan 6, 2013.
			'backup_mode'			=>		$profile['backup_mode'],						// Tells whether modern or classic mode.
			'type'					=>		$type,											// db, full, or export.
			'profile'				=>		$profile,										// Backup profile data.
			'default_profile'		=>		pb_mainwp_backupbuddy::$options['profiles'][0],		// Default profile.
			'start_time'			=>		time(),											// When backup started. Now.
			'finish_time'			=>		0,
			'updated_time'			=>		time(),											// When backup last updated. Subsequent steps update this.
			'status'				=>		array(),										// TODO: what goes in this?
			'max_execution_time'	=>		mainwp_backupbuddy_core::adjustedMaxExecutionTime(),	// Max execution time for chunking, taking into account user-specified override in settings (if any).
			'archive_size'			=>		0,
			'schedule_title'		=>		$schedule_title,								// Title of the schedule that made this backup happen (if applicable).
			'backup_directory'		=>		mainwp_backupbuddy_core::getBackupDirectory(),			// Directory backups stored in.
			'archive_file'			=>		$archiveFile,									// Unique backup ZIP filename.
			'archive_url'			=>		$archiveURL,									// Target download URL.
			'trigger'				=>		$trigger,										// How backup was triggered: manual or scheduled.
			'zip_method_strategy'	=>		pb_mainwp_backupbuddy::$options['zip_method_strategy'],// Enumerated zip method strategy
			'compression'			=>		$compression, 									//$profile['compression'], // Boolean - future enumerated?
			'ignore_zip_warnings'	=>		pb_mainwp_backupbuddy::$options['ignore_zip_warnings'],// Boolean - future bitmask?
			'ignore_zip_symlinks'	=>		pb_mainwp_backupbuddy::$options['ignore_zip_symlinks'],// Boolean - future bitmask?
			'steps'					=>		array(),										// Backup steps to perform. Set next in this code.
			'integrity'				=>		array(),										// Used later for tests and stats post backup.
			'temp_directory'		=>		'',												// Temp directory to store SQL and DAT file. Differs for exports. Defined in a moment...
			'backup_root'			=>		'',												// Where to start zipping from. Usually root of site. Defined in a moment...
			'export_plugins'		=>		array(),										// Plugins to export during MS export of a subsite.
			'additional_table_includes'	=>	array(),
			'additional_table_excludes'	=>	array(),
			'directory_exclusions'		=>	mainwp_backupbuddy_core::get_directory_exclusions( $profile, false, $serial ), // Do not trim trailing slash
			'table_sizes'			=>		array(),										// Array of tables to backup AND their sizes.
			'breakout_tables'		=>		array(),										// Array of tables that will be broken out to separate steps.
			'force_single_db_file'	=>		$forceSingleDatabaseFile,						// Whether forcing to a single db_1.sql file.
			'deployment_log'		=>		'',												// PUSH: URL to the importtbuddy status log for deployments. PULL: serial for the backup status log to retrieve via remote api.
			'deployment_direction'	=>		$deployDirection,								// Deployment direction, if any. valid: '', push, pull.
			'deployment_destination' =>		$deployDestinationSettings,						// Deployment remote destination settings if deployment.
			'runnerUID'				=>		get_current_user_id(),							// UID of whomever is running this backup. 0 if scheduled or ran by other automation means.
			'wp-config_in_parent'	=>		false,
		);
		pb_mainwp_backupbuddy::status( 'startSubFunction', json_encode( array( 'function' => 'file_excludes', 'title' => 'Found ' . count( $this->_backup['directory_exclusions'] ) . ' file or directory exclusions.' ) ) );
		
		// Warn if excluding key paths.
		$alertFileExcludes = mainwp_backupbuddy_core::alert_core_file_excludes( $this->_backup['directory_exclusions'] );
		foreach( $alertFileExcludes as $alertFileExcludeId => $alertFileExclude ) {
			pb_mainwp_backupbuddy::status( 'warning', $alertFileExclude );
		}
		
		pb_mainwp_backupbuddy::anti_directory_browsing( mainwp_backupbuddy_core::getTempDirectory(), $die = false );
		
		// Figure out paths.
		if ( ( $this->_backup['type'] == 'full' ) || ( $this->_backup['type'] == 'files' ) ) {
			$this->_backup['temp_directory'] = mainwp_backupbuddy_core::getTempDirectory() . $serial . '/';
			$this->_backup['backup_root'] = ABSPATH; // ABSPATH contains trailing slash.
		} elseif ( $this->_backup['type'] == 'db' ) {
			$this->_backup['temp_directory'] = mainwp_backupbuddy_core::getTempDirectory() . $serial . '/';
			$this->_backup['backup_root'] = $this->_backup['temp_directory'];
		} elseif ( $this->_backup['type'] == 'export' ) {
			// WordPress unzips into wordpress subdirectory by default so must include that in path.
			$this->_backup['temp_directory'] = mainwp_backupbuddy_core::getTempDirectory() . $serial . '/wordpress/wp-content/uploads/backupbuddy_temp/' . $serial . '/'; // We store temp data for export within the temporary WordPress installation within the temp directory. A bit confusing; sorry about that.
			$this->_backup['backup_root'] = mainwp_backupbuddy_core::getTempDirectory() . $serial . '/wordpress/';
		} else {
			pb_mainwp_backupbuddy::status( 'error', __('Backup FAILED. Unknown backup type.', 'it-l10n-mainwp-backupbuddy' ) );
			pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
		}
		pb_mainwp_backupbuddy::status( 'details', 'Temp directory: `' . $this->_backup['temp_directory'] . '`.' );
		pb_mainwp_backupbuddy::status( 'details', 'Backup root: `' . $this->_backup['backup_root'] . '`.' );
		
		
		// Plugins to export (only for MS exports).
		if ( count( $export_plugins ) > 0 ) {
			$this->_backup['export_plugins'] = $export_plugins;
		}
		
		
		// Calculate additional database table inclusion/exclusion.
		$this->_backup['additional_table_includes'] = mainwp_backupbuddy_core::get_mysqldump_additional( 'includes', $profile );
		$this->_backup['additional_table_excludes'] = mainwp_backupbuddy_core::get_mysqldump_additional( 'excludes', $profile );
		
		// Does wp-config.php exist in parent dir instead of normal location?
		if ( file_exists( ABSPATH . 'wp-config.php' ) ) { // wp-config in normal place.
			pb_mainwp_backupbuddy::status( 'details', 'wp-config.php found in normal location.' );
		} elseif ( @file_exists( dirname( ABSPATH ) . '/wp-config.php' ) ) { // Config in parent.
			pb_mainwp_backupbuddy::status( 'message', 'wp-config.php found in parent directory. Copying wp-config.php to temporary location for backing up.' );
			$this->_backup['wp-config_in_parent'] = true;
			$wp_config_parent = true;
		} else {
			pb_mainwp_backupbuddy::status( 'error', 'Error #3289493: wp-config.php not found in any normal location. Check file permissions. open_basedir() restrictions may block readability of wp-config.php in parent directory.' );
		}
		
		/********* Begin setting up steps array. *********/
		
		if ( $type == 'export' ) {
			pb_mainwp_backupbuddy::status( 'details', 'Setting up export-specific steps.' );
			
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_download_extract_wordpress',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_create_wp_config',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_copy_plugins',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_copy_themes',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_copy_media',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->_backup['steps'][] = array(
				'function'		=>	'ms_copy_users_table', // Create temp user and usermeta tables.
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
		}
		
		if ( ( 'pull' != $deployDirection ) && ( '1' != $profile['skip_database_dump'] ) && ( $profile['type'] != 'files' ) ) { // Backup database if not skipping AND not a files only backup.
			
			global $wpdb;
			// Default tables to backup.
			if ( $type == 'export' ) { // Multisite Subsite export only dumps tables specific to this subsite prefix.
				$base_dump_mode = 'prefix';
			} else { // Non-multisite export so use profile to determine tables to backup.
				if ( $profile['backup_nonwp_tables'] == '1' ) { // Backup all tables.
					$base_dump_mode = 'all';
				} elseif ( $profile['backup_nonwp_tables'] == '2' ) { // Backup no tables by default. Relies on listed additional tables.
					$base_dump_mode = 'none';
				} else { // Only backup matching prefix.
					$base_dump_mode = 'prefix';
				}
			}
			
			$additional_tables = $this->_backup['additional_table_includes'];
			if ( $type == 'export' ) {
				global $wpdb;
				array_push( $additional_tables, $wpdb->prefix . "users" );
				array_push( $additional_tables, $wpdb->prefix . "usermeta" );
			}
			
			// Warn if excluding key WP tables.
			$tableExcludes = mainwp_backupbuddy_core::alert_core_table_excludes( $this->_backup['additional_table_excludes'] );
			foreach( $tableExcludes as $tableExcludeId => $tableExclude ) {
				pb_mainwp_backupbuddy::status( 'warning', $tableExclude );
			}
			
			// Calculate tables to dump based on the provided information. $tables will be an array of tables.
			$tables = mainwp_backupbuddy_core::calculate_tables( $base_dump_mode, $additional_tables, $this->_backup['additional_table_excludes'] );
			pb_mainwp_backupbuddy::status( 'startSubFunction', json_encode( array( 'function' => 'calculate_tables', 'title' => 'Found ' . count( $tables ) . ' tables to backup based on settings.', 'more' => 'Tables: ' . implode( ', ', $tables ) ) ) );
			
			// If calculations show NO database tables should be backed up then change mode to skip database dump.
			if ( 0 == count( $tables ) ) {
				pb_mainwp_backupbuddy::status( 'warning', 'WARNING #857272: No database tables will be backed up based on current settings. This will not be a complete backup. Adjust settings if this is not intended and use with caution. Skipping database dump step.' );
				$profile['skip_database_dump'] = '1';
				$this->_backup['profile']['skip_database_dump'] = '1';
			} else { // One or more tables set to backup.
				
				// Obtain tables sizes. Surround each table name by a single quote and implode with commas for SQL query to get sizes.
				$tables_formatted = array();
				foreach( $tables as $key => $formatted ) {
					$tables_formatted[$key] = "'" . $formatted . "'";
				}
				$tables_formatted = implode( ',', $tables_formatted );
				$sql = "SHOW TABLE STATUS WHERE Name IN({$tables_formatted});";
				$rows = $wpdb->get_results( $sql, ARRAY_A );
				if ( false === $rows ) {
					pb_mainwp_backupbuddy::alert( 'Error #85473474: Unable to retrieve table status. Query: `' . $sql . '`.', true );
					return false;
				}
				$totalDatabaseSize = 0;
				foreach( $rows as $row ) {
					$this->_backup['table_sizes'][ $row['Name'] ] = ( $row['Data_length'] + $row['Index_length'] );
					$totalDatabaseSize += $this->_backup['table_sizes'][ $row['Name'] ];
				}
				unset( $rows );
				unset( $tables_formatted );
				
				$databaseSize = pb_mainwp_backupbuddy::$format->file_size( $totalDatabaseSize );
				pb_mainwp_backupbuddy::status( 'details', 'Total calculated database size: `' . $databaseSize . '`.' );
				
				// Step through tables we want to break out and figure out which ones were indeed set to be backed up and break them out.
				if ( pb_mainwp_backupbuddy::$options['breakout_tables'] == '0' ) { // Breaking out DISABLED.
					pb_mainwp_backupbuddy::status( 'details', 'Breaking out tables DISABLED based on settings.' );
				} else { // Breaking out ENABLED.
					// Tables we will try to break out into standalone steps if possible.
					$breakout_tables_defaults = array(
						$wpdb->prefix . 'posts',
						$wpdb->prefix . 'postmeta',
					);
					
					pb_mainwp_backupbuddy::status( 'details', 'Breaking out tables ENABLED based on settings. Tables to be broken out into individual steps: `' . implode( ', ', $breakout_tables_defaults ) . '`.' );
					foreach( (array)$breakout_tables_defaults as $breakout_tables_default ) {
						if ( in_array( $breakout_tables_default, $tables ) ) {
							$this->_backup['breakout_tables'][] = $breakout_tables_default;
							$tables = array_diff( $tables, array( $breakout_tables_default ) ); // Remove from main table backup list.
						}
					}
					unset( $breakout_tables_defaults ); // No longer needed.
				}
				
				
				$this->_backup['steps'][] = array(
					'function'		=>	'backup_create_database_dump',
					'args'			=>	array( $tables ),
					'start_time'	=>	0,
					'finish_time'	=>	0,
					'attempts'		=>	0,
				);
				
				// Set up backup steps for additional broken out tables.
				foreach( (array)$this->_backup['breakout_tables'] as $breakout_table ) {
					$this->_backup['steps'][] = array(
						'function'		=>	'backup_create_database_dump',
						'args'			=>	array( array( $breakout_table ) ),
						'start_time'	=>	0,
						'finish_time'	=>	0,
						'attempts'		=>	0,
					);
				}
				
			} // end there being tables to backup.
			
		} else {
			pb_mainwp_backupbuddy::status( 'message', __( 'Skipping database dump based on settings / profile type.', 'it-l10n-mainwp-backupbuddy' ) . ' Backup type: `' . $type . '`.' );
		}
		
		if ( 'pull' != $deployDirection ) {
			$this->_backup['steps'][] = array(
				'function'		=>	'backup_zip_files',
				'args'			=>	array(),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			
			if ( $type == 'export' ) {
				$this->_backup['steps'][] = array( // Multisite export specific cleanup.
					'function'		=>	'ms_cleanup', // Removes temp user and usermeta tables.
					'args'			=>	array(),
					'start_time'	=>	0,
					'finish_time'	=>	0,
					'attempts'		=>	0,
				);
			}
			
			if ( $profile['integrity_check'] == '1' ) {
				pb_mainwp_backupbuddy::status( 'details', __( 'Integrity check will be performed based on settings for this profile.', 'it-l10n-mainwp-backupbuddy' ) );
				$this->_backup['steps'][] = array(
					'function'		=>	'integrity_check',
					'args'			=>	array(),
					'start_time'	=>	0,
					'finish_time'	=>	0,
					'attempts'		=>	0,
				);
			} else {
				pb_mainwp_backupbuddy::status( 'details', __( 'Skipping integrity check step based on settings for this profile.', 'it-l10n-mainwp-backupbuddy' ) );
			}
		}
		
		$this->_backup['steps'][] = array(
			'function'		=>	'post_backup',
			'args'			=>	array(),
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		
		// Prepend and append pre backup and post backup steps.				
		$this->_backup['steps'] = array_merge( $pre_backup, $this->_backup['steps'], $post_backup );
		
		/********* End setting up steps array. *********/
		
		
		// Save what we have so far so that any errors below will end up displayed to user.
		$this->_backup_options->save();
		
		
		/********* Begin directory creation and security. *********/
		
		pb_mainwp_backupbuddy::anti_directory_browsing( mainwp_backupbuddy_core::getBackupDirectory() );
		
		// Prepare temporary directory for holding SQL and data file.
		if ( mainwp_backupbuddy_core::getTempDirectory() == '' ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #54534344. Temp directory blank. Please deactivate then reactivate plugin to reset.' );
			return false;
		}
		
		if ( !file_exists( $this->_backup['temp_directory'] ) ) {
			if ( pb_mainwp_backupbuddy::$filesystem->mkdir( $this->_backup['temp_directory'] ) === false ) {
				pb_mainwp_backupbuddy::status( 'error', 'Error #9002b. Unable to create temporary storage directory (' . $this->_backup['temp_directory'] . ')' );
				return false;
			}
		}
		if ( !is_writable( $this->_backup['temp_directory'] ) ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #9015. Temp data directory is not writable. Check your permissions. (' . $this->_backup['temp_directory'] . ')' );
			return false;
		}
		pb_mainwp_backupbuddy::anti_directory_browsing( ABSPATH . 'wp-content/uploads/backupbuddy_temp/' );
		
		// Prepare temporary directory for holding ZIP file while it is being generated.
		$this->_backup['temporary_zip_directory'] = mainwp_backupbuddy_core::getBackupDirectory() . 'temp_zip_' . $this->_backup['serial'] . '/';
		if ( !file_exists( $this->_backup['temporary_zip_directory'] ) ) {
			if ( pb_mainwp_backupbuddy::$filesystem->mkdir( $this->_backup['temporary_zip_directory'] ) === false ) {
				pb_mainwp_backupbuddy::status( 'details', 'Error #9002c. Unable to create temporary ZIP storage directory (' . $this->_backup['temporary_zip_directory'] . ')' );
				return false;
			}
		}
		if ( !is_writable( $this->_backup['temporary_zip_directory'] ) ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #9015. Temp data directory is not writable. Check your permissions. (' . $this->_backup['temporary_zip_directory'] . ')' );
			return false;
		}
		
		/********* End directory creation and security *********/
		
		
		// Generate backup DAT (data) file containing details about the backup.
		if ( $this->backup_create_dat_file( $trigger ) !== true ) {
			pb_mainwp_backupbuddy::status( 'details', __('Problem creating DAT file.', 'it-l10n-mainwp-backupbuddy' ) );
			return false;
		}
		
		// Generating ImportBuddy file to include in the backup for FULL BACKUPS ONLY currently. Cannot put in DB because it would be in root and be excluded or conflict on extraction.
		if ( $type == 'full' ) {
			if ( pb_mainwp_backupbuddy::$options['include_importbuddy'] == '1' ) {
				pb_mainwp_backupbuddy::status( 'details', 'Generating ImportBuddy tool to include in backup archive: `' . $this->_backup['temp_directory'] . 'importbuddy.php`.' );
				pb_mainwp_backupbuddy::status( 'startAction', 'importbuddyCreation' );
				mainwp_backupbuddy_core::importbuddy( $this->_backup['temp_directory'] . 'importbuddy.php' );
				pb_mainwp_backupbuddy::status( 'finishAction', 'importbuddyCreation' );
				pb_mainwp_backupbuddy::status( 'details', 'ImportBuddy generation complete.' );
			} else { // dont include importbuddy.
				pb_mainwp_backupbuddy::status( 'details', 'ImportBuddy tool inclusion in ZIP backup archive skipped based on settings or backup type.' );
			}
		}
		
		// Save all of this.
		$this->_backup['init_complete'] = true; // pre_backup() completed.
		$this->_backup_options->save();
		
		
		pb_mainwp_backupbuddy::status( 'details', __('Finished pre-backup procedures.', 'it-l10n-mainwp-backupbuddy' ) );
		pb_mainwp_backupbuddy::status( 'milestone', 'finish_settings' );
		
		pb_mainwp_backupbuddy::status( 'finishFunction', json_encode( array( 'function' => 'pre_backup' ) ) );
		return true;
		
	} // End pre_backup().
	
	
	
	/*	process_backup()
	 *	
	 *	Process and run the next backup step.
	 *	
	 *	@param		string		$serial		Unique backup identifier.
	 *	@param		string		$trigger	What triggered this processing: manual or scheduled.
	 *	@return		boolean					True on success, false otherwise.
	 */
	function process_backup( $serial, $trigger = 'manual' ) {
				
		
	} // End process_backup().
	
	
	/*	backup_create_dat_file()
	 *	
	 *	Generates backupbuddy_dat.php within the temporary directory containing the
	 *	random serial in its name. This file contains a serialized array that has been
	 *	XOR encrypted for security.  The XOR key is backupbuddy_SERIAL where SERIAL
	 *	is the randomized set of characters in the ZIP filename. This file contains
	 *	various information about the source site.
	 *	
	 *	@param		string			$trigger			What triggered this backup. Valid values: scheduled, manual.
	 *	@return		boolean			true on success making dat file; else false
	 */
	function backup_create_dat_file( $trigger ) {
		
		$settings = array(
			'start_time'			=> $this->_backup['start_time'],
			'backup_type'			=> $this->_backup['type'],
			'profile'				=> $this->_backup['profile'],
			'serial'				=> $this->_backup['serial'],
			'breakout_tables'		=> $this->_backup['breakout_tables'],
			'table_sizes'			=> $this->_backup['table_sizes'],
			'force_single_db_file'	=> $this->_backup['force_single_db_file'],
			'deployment_direction'	=> $this->_backup['deployment_direction'],
			'trigger'				=> $this->_backup['trigger'],
			'skip_database_dump'	=> $this->_backup['profile']['skip_database_dump'],
			'db_excludes'			=> mainwp_backupbuddy_core::get_mysqldump_additional( 'excludes', $this->_backup['profile'] ),
			'db_includes'			=> mainwp_backupbuddy_core::get_mysqldump_additional( 'includes', $this->_backup['profile'] ),
		);
		
		$dat_file = $this->_backup['temp_directory'] . 'backupbuddy_dat.php';
		$dat_content = mainwp_backupbuddy_core::render_dat_contents( $settings, $dat_file );
		if ( ! is_array( $dat_content ) ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #34894834: Unable to render DAT file. Check permissions. Details: `' . $dat_content . '`. Fatal error.' );
			return false;
		}
		
		// Handle wp-config.php file in a parent directory, copying to temp location.
		if ( true === $dat_content['wp-config_in_parent'] ) {
			if ( true === copy( dirname( ABSPATH ) . '/wp-config.php', $this->_backup['temp_directory'] . 'wp-config.php' ) ) {
				pb_mainwp_backupbuddy::status( 'details', 'Copied wp-config.php from parent directory into working temp directory for backup.' );
			} else {
				pb_mainwp_backupbuddy::status( 'error', 'Error #82394474: Failed to copy wp-config.php from parent directory into working temp directory for backup.' );
			}
		}
		
		pb_mainwp_backupbuddy::status( 'details', __('Finished creating DAT (data) file.', 'it-l10n-mainwp-backupbuddy' ) );
		
		return true;
		
	} // End backup_create_dat_file().
	
	
	
	/*	backup_create_database_dump()
	 *	
	 *	Prepares configuration and passes to the mysqlbuddy library to handle backing up the database.
	 *	Automatically handles falling back to compatibility modes.
	 *	
	 *	@return		boolean				True on success; false otherwise.
	 */
	function backup_create_database_dump( $tables, $rows_start = 0 ) {
		
		pb_mainwp_backupbuddy::status( 'milestone', 'start_database' );
		pb_mainwp_backupbuddy::status( 'message', __('Starting database backup process.', 'it-l10n-mainwp-backupbuddy' ) );
		
		if ( 'php' == pb_mainwp_backupbuddy::$options['database_method_strategy'] ) {
			$force_methods = array( 'php' );
		} elseif ( 'commandline' == pb_mainwp_backupbuddy::$options['database_method_strategy'] ) {
			$force_methods = array( 'commandline' );
		} elseif ( 'all' == pb_mainwp_backupbuddy::$options['database_method_strategy'] ) {
			$force_methods = array( 'php', 'commandline' );
		} else {
			pb_mainwp_backupbuddy::status( 'error', 'Error #48934: Invalid forced database dump method setting: `' . pb_mainwp_backupbuddy::$options['database_method_strategy'] . '`.' );
			return false;
		}
		
		$maxExecution = mainwp_backupbuddy_core::adjustedMaxExecutionTime();
		if ( $this->_backup['profile']['backup_mode'] == '1' ) { // Disable DB chunking when in classic mode.
			$maxExecution = -1;
		}
		
		// Load mysqlbuddy and perform dump.
		require_once( pb_mainwp_backupbuddy::plugin_path() . '/lib/mysqlbuddy/mysqlbuddy.php' );
		global $wpdb;
		pb_mainwp_backupbuddy::$classes['mysqlbuddy'] = new pb_mainwp_backupbuddy_mysqlbuddy( DB_HOST, DB_NAME, DB_USER, DB_PASSWORD, $wpdb->prefix, $force_methods, $maxExecution ); // $database_host, $database_name, $database_user, $database_pass, $old_prefix, $force_method = array()
		
		// Force to single db_1.sql file if enabled via advanced options.
		if ( '1' == pb_mainwp_backupbuddy::$options['force_single_db_file'] ) {
			pb_mainwp_backupbuddy::$classes['mysqlbuddy']->force_single_db_file( true );
		}
		
		// Do the database dump.
		$result = pb_mainwp_backupbuddy::$classes['mysqlbuddy']->dump( $this->_backup['temp_directory'], $tables, $rows_start ); // if array, returns tables,rowstart
		
		if ( is_array( $result ) ) { // Chunking.
			$newStep = array(
				'function'		=>	'backup_create_database_dump',
				'args'			=>	array( $result[0], $result[1] ),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			//error_log( print_r( $this->_backup_options->options['steps'], true ) );
			array_splice( $this->_backup_options->options['steps'], $this->_currentStepIndex + 1, 0, array( $newStep ) );
			//error_log( print_r( $this->_backup_options->options['steps'], true ) );
			$this->_backup_options->save();
			pb_mainwp_backupbuddy::status( 'details', 'Inserted additional database dump step at `' . ( $this->_currentStepIndex + 1 ) . '` to resume at row `' . $result[1] . '`. The next chunk will proceed shortly.' );
		}
		
		// Made it this far so not chunking at this point.
		
		// Check and make sure mysql server is still around. If it's missing at this point we may not be able to trust that it succeeded properly.
		/*
		// REMOVED 3-3-2014. PHP dump now checks connection status after each table and command line is not related to this. Settings are all stored in fileoptions so this is no longer relevant.
		
		global $wpdb;
		if ( @mysql_ping( $wpdb->dbh ) === false ) { // No longer connected to database if false.
			pb_mainwp_backupbuddy::status( 'error', __( 'ERROR #9027b: The mySQL server went away at some point during the database dump step. This is almost always caused by mySQL running out of memory or the mysql server timing out far too early. Contact your host. The database dump integrity can no longer be guaranteed so the backup has been halted.' ) );
			if ( $result === true ) {
				pb_mainwp_backupbuddy::status( 'details', 'The database dump reported SUCCESS prior to this problem.' );
			} else {
				pb_mainwp_backupbuddy::status( 'details', 'The database dump reported FAILURE prior to this problem.' );
			}
			pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
			return false;
		}
		*/
		
		
		return $result;

	} // End backup_create_database_dump().
	
	
	
	/*	backup_zip_files()
	 *	
	 *	Create ZIP file containing everything.
	 *	Currently this is just a wrapper around zip system specific functions
	 *	
	 *	@return		boolean			True on success; false otherwise.
	 */
	function backup_zip_files( $state = array() ) {
	
		
		
	} // End backup_zip_files().
	
	/*	trim_old_archives()
	 *	
	 *	Get rid of excess archives based on user-defined parameters.
	 *	!!!!! IMPORTANT: The order of application of archivel limiting is important. Users have configured their settings based on the defined order on the Settings page. !!!!!
	 *	
	 *	@param		
	 *	@return		
	 */
	function trim_old_archives() {
		
		pb_mainwp_backupbuddy::status( 'details', __('Trimming old archives (if needed).', 'it-l10n-mainwp-backupbuddy' ) );
		
		$summed_size = 0;
		
		$file_list = glob( mainwp_backupbuddy_core::getBackupDirectory() . 'backup*.zip' );
		if ( is_array( $file_list ) && !empty( $file_list ) ) {
			foreach( (array) $file_list as $file ) {
				$file_stats = stat( $file );
				$modified_time = $file_stats['mtime'];
				$filename = str_replace( mainwp_backupbuddy_core::getBackupDirectory(), '', $file ); // Just the file name.
				$files[$modified_time] = array(
													'filename'				=>		$filename,
													'size'					=>		$file_stats['size'],
													'modified'				=>		$modified_time,
												);
				$summed_size += ( $file_stats['size'] / 1048576 ); // MB
			}
		}
		unset( $file_list );
		if ( empty( $files ) ) { // return if no archives (nothing else to do).
			pb_mainwp_backupbuddy::status( 'details', __( 'No old archive trimming needed.', 'it-l10n-mainwp-backupbuddy' ) );
			return true;
		} else {
			krsort( $files );
		}
		
		$trim_count = 0;
		
		
		// Limit by age if set.
		if ( (int) pb_mainwp_backupbuddy::$options['archive_limit_age'] > 0 ) {
			foreach( $files as $file_modified => $file ) {
				if ( ! is_numeric( $file['modified'] ) ) { // Could not get age so skipping.
					continue;
				}
				$backup_age = (int) ( time() - $file['modified'] ) / 60 / 60 / 24;
				if ( $backup_age > pb_mainwp_backupbuddy::$options['archive_limit_age'] ) { // Too old; delete!
					pb_mainwp_backupbuddy::status( 'details', __('Deleting old archive `' . $file['filename'] . '` as it exceeds the maximum age limit `' . pb_mainwp_backupbuddy::$options['archive_limit_age'] . '` allowed at `' . $backup_age . '` days.' ) );
					unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
					unset( $files[$file_modified] );
					$trim_count++;
				} else {
					//pb_mainwp_backupbuddy::status( 'details', 'Not deleted: ' . $backup_age );
				}
			}
		} // end age limit.
		
		
		// Limit by number of archives by backup type FULL.
		if ( pb_mainwp_backupbuddy::$options['archive_limit_full'] > 0 ) {
			// MAY need to trim.
			$i = 0;
			foreach( $files as $file_modified => $file ) {
				if ( false !== strpos( $file['filename'], '-full-' ) ) {
					$i++;
					if ( $i > pb_mainwp_backupbuddy::$options['archive_limit_full'] ) {
						pb_mainwp_backupbuddy::status( 'details', sprintf( __('Deleting old archive `%s` as it causes archives to exceed total number of FULL backups allowed.', 'it-l10n-mainwp-backupbuddy' ), $file['filename'] ) );
						unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
						unset( $files[$file_modified] );
						$trim_count++;
					}
				}
			}
		} // end number of archives of type FULL limit.
		
		
		// Limit by number of archives by backup type DB.
		if ( pb_mainwp_backupbuddy::$options['archive_limit_db'] > 0 ) {
			// MAY need to trim.
			$i = 0;
			foreach( $files as $file_modified => $file ) {
				if ( false !== strpos( $file['filename'], '-db-' ) ) {
					$i++;
					if ( $i > pb_mainwp_backupbuddy::$options['archive_limit_db'] ) {
						pb_mainwp_backupbuddy::status( 'details', sprintf( __('Deleting old archive `%s` as it causes archives to exceed total number of DATABASE backups allowed.', 'it-l10n-mainwp-backupbuddy' ), $file['filename'] ) );
						unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
						unset( $files[$file_modified] );
						$trim_count++;
					}
				}
			}
		} // end number of archives of type DB limit.
		
		
		// Limit by number of archives by backup type FILES.
		if ( pb_mainwp_backupbuddy::$options['archive_limit_files'] > 0 ) {
			// MAY need to trim.
			$i = 0;
			foreach( $files as $file_modified => $file ) {
				if ( false !== strpos( $file['filename'], '-files-' ) ) {
					$i++;
					if ( $i > pb_mainwp_backupbuddy::$options['archive_limit_files'] ) {
						pb_mainwp_backupbuddy::status( 'details', sprintf( __('Deleting old archive `%s` as it causes archives to exceed total number of FILES backups allowed.', 'it-l10n-mainwp-backupbuddy' ), $file['filename'] ) );
						unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
						unset( $files[$file_modified] );
						$trim_count++;
					}
				}
			}
		} // end number of archives of type FILES limit.
		
		
		// Limit by number of archives if set. Deletes oldest archives over this limit.
		if ( ( pb_mainwp_backupbuddy::$options['archive_limit'] > 0 ) && ( count( $files ) ) > pb_mainwp_backupbuddy::$options['archive_limit'] ) {
			// Need to trim.
			$i = 0;
			foreach( $files as $file_modified => $file ) {
				$i++;
				if ( $i > pb_mainwp_backupbuddy::$options['archive_limit'] ) {
					pb_mainwp_backupbuddy::status( 'details', sprintf( __('Deleting old archive `%s` as it causes archives to exceed total number allowed.', 'it-l10n-mainwp-backupbuddy' ), $file['filename'] ) );
					unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
					unset( $files[$file_modified] );
					$trim_count++;
				}
			}
		} // end number of archives limit.
		
		
		// Limit by size of archives, oldest first if set.
		$files = array_reverse( $files, true ); // Reversed so we delete oldest files first as long as size limit still is surpassed; true = preserve keys.
		
		if ( 0 == pb_mainwp_backupbuddy::$options['archive_limit_size'] ) { // Limit of 0. Use BIG limit.
			$lesser_limit = pb_mainwp_backupbuddy::$options['archive_limit_size_big'];
		} elseif ( 0 == pb_mainwp_backupbuddy::$options['archive_limit_size_big'] ) { // Big is zero. Use normal limit.
			$lesser_limit = pb_mainwp_backupbuddy::$options['archive_limit_size'];
		} else { // Big is set so decide which is smaller.
			$lesser_limit = min( pb_mainwp_backupbuddy::$options['archive_limit_size'], pb_mainwp_backupbuddy::$options['archive_limit_size_big'] );
		}
		
		if ( ( $lesser_limit > 0 ) && ( $summed_size > $lesser_limit ) ) { // A limit was found and we exceed it.
			// Need to trim.
			foreach( $files as $file_modified => $file ) {
				if ( $summed_size > $lesser_limit ) {
					$summed_size = $summed_size - ( $file['size'] / 1048576 );
					pb_mainwp_backupbuddy::status( 'details', sprintf( __('Deleting old archive `%s` due as it causes archives to exceed total size allowed.', 'it-l10n-mainwp-backupbuddy' ),  $file['filename'] ) );
					if ( $file['filename'] != basename( $this->_backup['archive_file'] ) ) { // Delete excess archives as long as it is not the just-made backup.
						unlink( mainwp_backupbuddy_core::getBackupDirectory() . $file['filename'] );
						unset( $files[$file_modified] );
						$trim_count++;
					} else {
						$message = __( 'ERROR #9028: Based on your backup archive limits (size limit) the backup that was just created would be deleted. Skipped deleting this backup. Please update your archive limits.' );
						pb_mainwp_backupbuddy::status( 'message', $message );
						mainwp_backupbuddy_core::mail_error( $message );
					}
				}
			}
		} // end combined file size limit.
		
		
		pb_mainwp_backupbuddy::status( 'details', 'Trimmed ' . $trim_count . ' old archives based on settings archive limits.' );
		return true;
		
	} // End trim_old_archives().
	
	
	
	/* integrity_check()
	 *
	 * Perform integrity check on backup file to confirm backup.
	 *
	 */
	function integrity_check() {
		
		return true;
		
	} // End integrity_check().
	
	
	
	/*	post_backup()
	 *	
	 *	Post-backup procedured. Clean up, send notifications, etc.
	 *	
	 *	@return		null
	 */
	function post_backup( $fail_mode = false, $cancel_backup = false ) {
		pb_mainwp_backupbuddy::status( 'message', __('Cleaning up after backup.', 'it-l10n-mainwp-backupbuddy' ) );
		
		// Delete temporary data directory.
		if ( file_exists( $this->_backup['temp_directory'] ) ) {
			pb_mainwp_backupbuddy::status( 'details', __('Removing temp data directory.', 'it-l10n-mainwp-backupbuddy' ) );
			pb_mainwp_backupbuddy::$filesystem->unlink_recursive( $this->_backup['temp_directory'] );
		}
		// Delete temporary ZIP directory.
		if ( file_exists( mainwp_backupbuddy_core::getBackupDirectory() . 'temp_zip_' . $this->_backup['serial'] . '/' ) ) {
			pb_mainwp_backupbuddy::status( 'details', __('Removing temp zip directory.', 'it-l10n-mainwp-backupbuddy' ) );
			pb_mainwp_backupbuddy::$filesystem->unlink_recursive( mainwp_backupbuddy_core::getBackupDirectory() . 'temp_zip_' . $this->_backup['serial'] . '/' );
		}
		
		
		if ( true === $fail_mode ) {
			pb_mainwp_backupbuddy::status( 'warning', 'Backup archive limiting has been skipped since there was an error to avoid deleting potentially good backups to make room for a potentially bad backup.' );
		} else {
			$this->trim_old_archives(); // Clean up any old excess archives pushing us over defined limits in settings.
		}
		
		
		if ( true === $cancel_backup ) {
			pb_mainwp_backupbuddy::status( 'details', 'Backup stopped so deleting backup ZIP file.' );
			$unlink_result = @unlink( $this->_backup['archive_file'] );
			if ( true === $unlink_result ) {
				pb_mainwp_backupbuddy::status( 'details', 'Deleted stopped backup file.' );
			} else {
				pb_mainwp_backupbuddy::status( 'error', 'Unable to delete stopped backup file. You should delete it manually as it may be damaged from stopping mid-backup. File to delete: `' . $this->_backup['archive_file'] . '`.' );
			}
			
			$this->_backup['finish_time'] = -1;
			//pb_mainwp_backupbuddy::save();
			$this->_backup_options->save();
			
		} else { // Not cancelled.
			$this->_backup['archive_size'] = @filesize( $this->_backup['archive_file'] );
			pb_mainwp_backupbuddy::status( 'details', __('Final ZIP file size', 'it-l10n-mainwp-backupbuddy' ) . ': ' . pb_mainwp_backupbuddy::$format->file_size( $this->_backup['archive_size'] ) );
			pb_mainwp_backupbuddy::status( 'archiveSize', pb_mainwp_backupbuddy::$format->file_size( $this->_backup['archive_size'] ) );
			
			if ( $fail_mode === false ) { // Not cancelled and did not fail so mark finish time.
				
				//error_log( print_r( $this->_backup_options->options, true ) );
				
				$archiveFile = basename( $this->_backup_options->options['archive_file'] );
				
				// Calculate backup download URL, if any.
				//$downloadURL = pb_mainwp_backupbuddy::ajax_url( 'download_archive' ) . '&backupbuddy_backup=' . $archiveFile;
				$downloadURL = '';
				$abspath = str_replace( '\\', '/', ABSPATH ); // Change slashes to handle Windows as we store backup_directory with Linux-style slashes even on Windows.
				$backup_dir = str_replace( '\\', '/', mainwp_backupbuddy_core::getBackupDirectory() );
				if ( FALSE !== stristr( $backup_dir, $abspath ) ) { // Make sure file to download is in a publicly accessible location (beneath WP web root technically).
					//pb_mainwp_backupbuddy::status( 'details', 'mydir: `' . $backup_dir . '`, abs: `' . $abspath . '`.');
					$sitepath = str_replace( $abspath, '', $backup_dir );
					$downloadURL = rtrim( site_url(), '/\\' ) . '/' . trim( $sitepath, '/\\' ) . '/' . $archiveFile;
				}
				
				
				$integrityIsOK = '-1';
				if ( isset( $this->_backup_options->options['integrity']['is_ok'] ) ) {
					$integrityIsOK = $this->_backup_options->options['integrity']['is_ok'];
				}
				
				$destinations = array();
				foreach( $this->_backup_options->options['steps'] as $step ) {
					if ( 'send_remote_destination' == $step['function'] ) {
						$destinations[] = array(
											'id' => $step['args'][0],
											'title' => pb_mainwp_backupbuddy::$options['remote_destinations'][ $step['args'][0] ]['title'],
											'type' => pb_mainwp_backupbuddy::$options['remote_destinations'][ $step['args'][0] ]['type'],
										);
					}
				}
				
				pb_mainwp_backupbuddy::status( 'details', 'Updating statistics for last backup completed and number of edits since last backup.' );
				$finishTime = microtime(true);
				pb_mainwp_backupbuddy::$options['last_backup_finish'] = $finishTime;
				pb_mainwp_backupbuddy::$options['last_backup_stats'] = array(
																	//'serial'			=> $this->_backup['serial'],
																	'archiveFile'		=> $archiveFile,
																	'archiveURL'		=> $downloadURL,
																	'archiveSize'		=> $this->_backup['archive_size'],
																	'start'				=> pb_mainwp_backupbuddy::$options['last_backup_start'],
																	'finish'			=> $finishTime,
																	'type'				=> $this->_backup_options->options['profile']['type'],
																	'profileTitle'		=> htmlentities( $this->_backup_options->options['profile']['title'] ),
																	'scheduleTitle'		=> $this->_backup_options->options['schedule_title'], // Empty string is no schedule.
																	'integrityStatus'	=> $integrityIsOK, // 1, 0, -1 (unknown)
																	'destinations'		=> $destinations, // Index is destination ID. Empty array if none.
																);
				//error_log( print_r( pb_mainwp_backupbuddy::$options['last_backup_stats'], true ) );
				pb_mainwp_backupbuddy::$options['edits_since_last'] = 0; // Reset edit stats for notifying user of how many posts/pages edited since last backup happened.
				pb_mainwp_backupbuddy::save();
			}
			
		}
		
		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/housekeeping.php' );
		mainwp_backupbuddy_housekeeping::cleanup_temp_dir();
		
		
		if ( $this->_backup['trigger'] == 'manual' ) {
			// Do nothing. No notifications as of pre-3.0 2012.
		} elseif ( $this->_backup['trigger'] == 'deployment' ) {
			// Do nothing. No notifications.
		} elseif ( $this->_backup['trigger'] == 'deployment_pulling' ) {
			// Do nothing.
		} elseif ( $this->_backup['trigger'] == 'scheduled' ) {
			if ( ( false === $fail_mode ) && ( false === $cancel_backup ) ) {
				pb_mainwp_backupbuddy::status( 'details', __( 'Sending scheduled backup complete email notification.', 'it-l10n-mainwp-backupbuddy' ) );
				$message = 'completed successfully in ' . pb_mainwp_backupbuddy::$format->time_duration( time() - $this->_backup['start_time'] ) . ".\n";
				mainwp_backupbuddy_core::mail_notify_scheduled( $this->_backup['serial'], 'complete', __( 'Scheduled backup', 'it-l10n-mainwp-backupbuddy' ) . ' "' . $this->_backup['schedule_title'] . '" ' . $message );
			}
		} else {
			pb_mainwp_backupbuddy::status( 'error', 'Error #4343434. Unknown backup trigger `' . $this->_backup['trigger'] . '`.' );
		}
		
		
		pb_mainwp_backupbuddy::status( 'message', __( 'Finished cleaning up.', 'it-l10n-mainwp-backupbuddy' ) );
		
		if ( true === $cancel_backup ) {
			pb_mainwp_backupbuddy::status( 'details', 'Backup cancellation complete.' );
			return false;
		} else {
			if ( true === $fail_mode ) {
				pb_mainwp_backupbuddy::status( 'details', __( 'As this backup did not pass the integrity check you should verify it manually or re-scan. Integrity checks can fail on good backups due to permissions, large file size exceeding memory limits, etc. You may manually disable integrity check on the Settings page but you will no longer be notified of potentially bad backups.', 'it-l10n-mainwp-backupbuddy' ) );
			} else {
				
				if ( ( $this->_backup['trigger'] != 'deployment' ) && ( $this->_backup['trigger'] != 'deployment_pulling' ) ) {
					//$stats = stat( $this->_backup['archive_file'] );
					//$sizeFormatted = pb_mainwp_backupbuddy::$format->file_size( $stats['size'] );
					pb_mainwp_backupbuddy::status(
						'archiveInfo',
						json_encode( array(
							'file' => basename( $this->_backup['archive_file'] ),
							'url' => pb_mainwp_backupbuddy::ajax_url( 'download_archive' ) . '&backupbuddy_backup=' . basename( $this->_backup['archive_file'] ),
							//'sizeBytes' => $stats,
							//'sizeFormatted' => $sizeFormatted,
						) )
					);
				}
				
			}
		}
		
		return true;
		
	} // End post_backup().
	
	
	
	/*	send_remote_destination()
	 *	
	 *	Send the current backup to a remote destination such as S3, Dropbox, FTP, etc.
	 *	Scheduled remote sends end up coming through here before passing to core.
	 *	
	 *	@param		int		$destination_id		Destination ID (remote destination array index) to send to.
	 *	@param		boolean	$delete_after		Whether or not to delete backup file after THIS successful remote transfer.
	 *	@return		boolean						Returns result of pb_mainwp_backupbuddy::send_remote_destination(). True (success) or false (failure).
	 */
	function send_remote_destination( $destination_id, $delete_after = false ) {
		
		pb_mainwp_backupbuddy::status( 'details', 'Sending file to remote destination ID: `' . $destination_id . '`. Delete after: `' . $delete_after . '`.' );
		pb_mainwp_backupbuddy::status( 'details', 'IMPORTANT: If the transfer is set to be chunked then only the first chunk status will be displayed during this process. Subsequent chunks will happen after this has finished.' );
		$response = mainwp_backupbuddy_core::send_remote_destination( $destination_id, $this->_backup['archive_file'], '', false, $delete_after );
		
		if ( false === $response ) { // Send failure.
			$error_message = 'BackupBuddy failed sending a backup to the remote destination "' . pb_mainwp_backupbuddy::$options['remote_destinations'][ $destination_id ]['title'] . '" (id: ' . $destination_id . '). Please verify and test destination settings and permissions. Check the error log for further details.';
			pb_mainwp_backupbuddy::status( 'error', 'Failure sending to remote destination. Details: ' . $error_message );
			mainwp_backupbuddy_core::mail_error( $error_message );
		}
		
	} // End send_remote_destination().
	
	
	
	// DEPRECATED Mar 5, 2013. - Dustin
	/*	post_remote_delete()
	 *	
	 *	Deletes backup archive. Used to delete the backup after sending to a remote destination for scheduled backups.
	 *	
	 *	@return		boolean		True on deletion success; else false.
	 */
	function post_remote_delete() {
		
		
		// DEPRECATED FUNCTION. DO NOT USE.
		
		
		pb_mainwp_backupbuddy::status( 'error', 'CALL TO DEPRECATED FUNCTION post_remote_delete().' );
		pb_mainwp_backupbuddy::status( 'details', 'Deleting local copy of file sent remote.' );
		if ( file_exists( $this->_backup['archive_file'] ) ) {
			unlink( $this->_backup['archive_file'] );
		}
		
		if ( file_exists( $this->_backup['archive_file'] ) ) {
			pb_mainwp_backupbuddy::status( 'details', __('Error. Unable to delete local archive as requested.', 'it-l10n-mainwp-backupbuddy' ) );
			return false; // Didnt delete.
		} else {
			pb_mainwp_backupbuddy::status( 'details', __('Deleted local archive as requested.', 'it-l10n-mainwp-backupbuddy' ) );
			return true; // Deleted.
		}
	} // End post_remote_delete().
	
	
	public function deploy_push_sendFile( $state, $sendFile, $sendPath, $sendType, $nextStep, $delete_after = false ) {
		$destination_settings = &pb_mainwp_backupbuddy::$options['remote_destinations'][ $state['destination_id'] ];
		$destination_settings['sendType'] = $sendType;
		$destination_settings['sendFilePath'] = $sendPath;
		$destination_settings['max_time'] = $state['minimumExecutionTime'];
		
		$identifier = $this->_backup['serial'] . '_' . md5( $sendFile . $sendType );
		if ( false === mainwp_backupbuddy_core::send_remote_destination( $state['destination_id'], $sendFile, 'Deployment', $send_importbuddy = false, $delete_after, $identifier, $destination_settings ) ) {
			$sendFile = ''; // Since failed just set file to blank so we can proceed to next without waiting.
		}
		
		$this->deploy_sendWait( $state, $sendFile, $sendPath, $sendType, $nextStep );
		return true;
	} // End deploy_push_sendFile().
	
	
	public function deploy_sendWait( $state, $sendFile, $sendPath, $sendType, $nextStep ) {
		$identifier = $this->_backup['serial'] . '_' . md5( $sendFile . $sendType );
		pb_mainwp_backupbuddy::status( 'details', 'Waiting for send to finish for file `' . $sendFile . '` with ID `' . $identifier . '`.' );
		
		$maxSendTime = 60*5;
		
		if ( '' == $sendFile ) { // File failed. Proceed to next.
			$this->insert_next_step( $nextStep );
			return true;
		}
		
		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/fileoptions.php' );
		pb_mainwp_backupbuddy::status( 'details', 'Fileoptions instance #38.' );
		$fileoptions_obj = new pb_mainwp_backupbuddy_fileoptions( mainwp_backupbuddy_core::getLogDirectory() . 'fileoptions/send-' . $identifier . '.txt', $read_only = false, $ignore_lock = true, $create_file = false );
		if ( true !== ( $result = $fileoptions_obj->is_ok() ) ) {
			pb_mainwp_backupbuddy::status( 'error', __('Fatal Error #9034 E. Unable to access fileoptions data.', 'it-l10n-mainwp-backupbuddy' ) . ' Error: ' . $result );
			pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
			return false;
		}
		pb_mainwp_backupbuddy::status( 'details', 'Fileoptions data loaded.' );
		$fileoptions = &$fileoptions_obj->options; // Set reference.
		if ( '0' == $fileoptions['finish_time'] ) { // Not finished yet. Insert next chunk to wait.
			$timeAgo = ( time() - $fileoptions['start_time'] );
			if ( $timeAgo > $maxSendTime) {
				pb_mainwp_backupbuddy::status( 'error', 'Error #4948348: Maximum allowed file send time of `' . $maxSendTime . '` seconds passed. Halting.' );
				pb_mainwp_backupbuddy::status( 'haltScript', '' ); // Halt JS on page.
			}
			pb_mainwp_backupbuddy::status( 'details', 'File send not yet finished. Started `' . $timeAgo . '` seconds ago. Inserting wait.' );
			
			$newStep = array(
				'function'		=>	'deploy_sendWait',
				'args'			=>	array( $state, $sendFile, $sendPath, $sendType, $nextStep ),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->insert_next_step( $newStep );
		} else { // Finished. Go to next step.
			$this->insert_next_step( $nextStep );
		}
		
		return true;
	} // End deploy_sendWait().
	
	
	
	/* deploy_push_start()
	 *
	 * Start a deployment of this site to a remote site.
	 *
	 */
	public function deploy_push_start( $state ) {
            
//		pb_mainwp_backupbuddy::status( 'details', 'Starting PUSH deployment process. Incoming state: `' . print_r( $state, true ) . '`.' );
//		pb_mainwp_backupbuddy::status( 'startSubFunction', json_encode( array( 'function' => 'deploy_push_start', 'title' => 'Found deployment.' ) ) );
//		
//		
//		
//		// Send backup zip. It will schedule its next chunks as needed. When done it calls the nextstep.
//		$nextStep = array(
//			'function'		=>	'deploy_push_sendContent',
//			'args'			=>	array( $state ),
//			'start_time'	=>	0,
//			'finish_time'	=>	0,
//			'attempts'		=>	0,
//		);
//		$this->deploy_push_sendFile( $state, $this->_backup['archive_file'], '', 'backup', $nextStep, $delete_after = true );
		
		return true;
	} // End deploy_push_start().
	
	
	
	/* deploy_pull_start()
	 *
	 * Start a deployment of this site to a remote site.
	 *
	 */
	public function deploy_pull_start( $state ) {
//		pb_mainwp_backupbuddy::status( 'details', 'Starting PULL deployment process. Incoming state: `' . print_r( $state, true ) . '`.' );
//		pb_mainwp_backupbuddy::status( 'startSubFunction', json_encode( array( 'function' => 'deploy_pull_start', 'title' => 'Found deployment.' ) ) );
//		
//		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/deploy.php' );
//		$deploy = new mainwp_backupbuddy_deploy( $state['destinationSettings'], $state );
//		
//		// If not pulling any DB contents then skip making remote backup file.
//		//error_log( print_r( $this->_backup['profile'], true ) );
//		if ( '1' == $this->_backup['profile']['skip_database_dump'] ) {
//			pb_mainwp_backupbuddy::status( 'details', 'No database tables selected for pulling.  Skipping remote database snapshot (backup) zip creation and inserting file pull step.' );
//			$newStep = array(
//				'function'		=>	'deploy_pull_files',
//				'args'			=>	array( $state ),
//				'start_time'	=>	0,
//				'finish_time'	=>	0,
//				'attempts'		=>	0,
//			);
//			$this->insert_next_step( $newStep );
//			return true;
//		} else {
//			pb_mainwp_backupbuddy::status( 'details', 'Database tables selected for pulling. Proceeding towards remote snapshot.' );
//		}
//		
//		// Get session token to restore so we won't be logged out. Place them in the remote backup profile array.
//		global $wpdb;
//		$sql = "SELECT meta_value FROM `" . DB_NAME . "`.`" . $wpdb->prefix . "usermeta` WHERE `user_id` = '" . $this->_backup['runnerUID'] . "' AND `meta_key` = 'session_tokens'";
//		pb_mainwp_backupbuddy::status( 'details', 'TokenSQL: ' . $sql );
//		$results = $wpdb->get_var( $sql );
//		$this->_backup['profile']['sessionTokens'] = unserialize( $results );
//		$this->_backup['profile']['sessionID'] = $this->_backup['runnerUID'];
//		//error_log( 'sourcetokens: ' . print_r( $this->_backup['profile'], true ) );
//		pb_mainwp_backupbuddy::status( 'details', 'Session tokens calculated.' );
//		
//		pb_mainwp_backupbuddy::status( 'details', 'About to remote call.' );
//		if ( false === ( $response = mainwp_backupbuddy_remote_api::remoteCall( $state['destination'], 'runBackup', array( 'profile' => base64_encode( json_encode( $this->_backup['profile'] ) ) ), $state['minimumExecutionTime'] ) ) ) {
//			pb_mainwp_backupbuddy::status( 'error', 'Error #44548985: Unable to start remote backup via remote API.' );
//			return false;
//		}
//		pb_mainwp_backupbuddy::status( 'details', 'Server response: `' . print_r( $response, true ) . '`.' );
//		$remoteBackupSerial = $response['backupSerial'];
//		$remoteBackupFile = $response['backupFile'];
//		$this->_backup_options->options['remote_backup_file'] = $remoteBackupFile;
//		pb_mainwp_backupbuddy::status( 'details', 'Remote backup file: `' . $this->_backup_options->options['remote_backup_file'] . '`.' );
//		$this->_backup_options->options['deployment_log'] = $remoteBackupSerial; // _getBackupStatus.php uses this serial for remote call retrival of the status log during the backup.
//		pb_mainwp_backupbuddy::status( 'details', 'Remote backup log: `' . $this->_backup_options->options['deployment_log'] . '`.' );
//		$this->_backup_options->save();
//		
//		pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to wait for backup to finish on remote server.' );
//		$newStep = array(
//			'function'		=>	'deploy_pull_runningBackup',
//			'args'			=>	array( $state ),
//			'start_time'	=>	0,
//			'finish_time'	=>	0,
//			'attempts'		=>	0,
//		);
//		$this->insert_next_step( $newStep );
		
		return true;
	} // End deploy_pull_start().
	
	
	
	public function deploy_pull_files( $state, $pullBackupArchive = '' ) {
		pb_mainwp_backupbuddy::status( 'details', 'Starting retrieval of files from source remote site.' );
		
		
		
		// Next step will be back here after any send file waiting finishes. Keep coming back here until there is nothing more to send.
		$nextStep = array(
			'function'		=>	'deploy_pull_files',
			'args'			=>	array(), // MUST populate state before passing off this next step.
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		
		// Count up number of files remaining.
		$mediaFileCount = 0;
		$pluginFileCount = 0;
		$themeFileCount = 0;
		$childThemeFileCount = 0;
		if ( true === $state['sendMedia'] ) {
			$mediaFileCount = count( $state['pullMediaFiles'] );
		}
		if ( count( $state['sendPlugins'] > 0 ) ) {
			$pluginFileCount = count( $state['pullPluginFiles'] );
		}
		if ( true === $state['sendTheme'] ) {
			$themeFileCount = count( $state['pullThemeFiles'] );
		}
		if ( true === $state['sendChildTheme'] ) {
			$childThemeFileCount = count( $state['pullChildThemeFiles'] );
		}
		
		$filesRemaining = $mediaFileCount + $pluginFileCount + $themeFileCount + $childThemeFileCount;
		if ( '' != $pullBackupArchive ) { // add in backup archive if not yet sent.
			$filesRemaining++;
		}
		pb_mainwp_backupbuddy::status( 'deployFilesRemaining', $filesRemaining );
		pb_mainwp_backupbuddy::status( 'details', 'Files remaining to retrieve: ' . $filesRemaining );
		
		if ( '' != $pullBackupArchive ) {
			$nextStep['args'] = array( $state );
			$state['pullLocalArchiveFile'] = ABSPATH . $pullBackupArchive;
			$nextStep['args'] = array( $state );
			return $this->deploy_getFile( $state, dirname( $state['pullLocalArchiveFile'] ) . '/', $pullBackupArchive, 'backup', $nextStep );
		}
		
		if ( true !== $state['sendMedia'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING pull of media files.' );
		} else {
			if ( $mediaFileCount > 0 ) { // Media files remain to send.
				$getFile = array_pop( $state['pullMediaFiles'] ); // Pop off last item in array. Faster than shift.
				$wp_upload_dir = wp_upload_dir();
				$nextStep['args'] = array( $state );
				pb_mainwp_backupbuddy::status( 'details', 'About to send media file.' );
				return $this->deploy_getFile( $state, $wp_upload_dir['basedir'] . '/', $getFile, 'media', $nextStep );
			}
		}
		
		if ( 0 == count( $state['sendPlugins'] ) ) {
			pb_mainwp_backupbuddy::status( 'details', 'No plugin files selected for transfer. Skipping send.' );
		} else {
			if ( $pluginFileCount > 0 ) { // Plugin files remain to send.
				pb_mainwp_backupbuddy::status( 'details', 'Plugins files remaining to send: ' . count( $state['pushPluginFiles'] ) );
				$getFile = array_pop( $state['pullPluginFiles'] ); // Pop off last item in array. Faster than shift.
				$pluginPath = wp_normalize_path( WP_PLUGIN_DIR ) . '/';
				$nextStep['args'] = array( $state );
				return $this->deploy_getFile( $state, $pluginPath, $getFile, 'plugin', $nextStep );
			}
		}
		
		if ( true !== $state['sendTheme'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING pull of theme files.' );
		} else {
			if ( $themeFileCount > 0 ) { // Plugin files remain to send.
				$getFile = array_pop( $state['pullThemeFiles'] ); // Pop off last item in array. Faster than shift.
				$themePath = get_template_directory(); // contains trailing slash.
				$nextStep['args'] = array( $state );
				return $this->deploy_getFile( $state, $themePath, $getFile, 'theme', $nextStep );
			}
		}
		
		if ( true !== $state['sendChildTheme'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING pull of child theme files.' );
		} else {
			if ( $childThemeFileCount > 0 ) { // Plugin files remain to send.
				$getFile = array_pop( $state['pullChildThemeFiles'] ); // Pop off last item in array. Faster than shift.
				$childThemePath = get_stylesheet_directory(); // contains trailing slash.
				$nextStep['args'] = array( $state );
				return $this->deploy_getFile( $state, $childThemePath, $getFile, 'childTheme', $nextStep );
			}
		}
		
		// If we made it here then all file sending is finished. Move on to next step.
		$nextStep = array(
			'function'		=>	'deploy_pull_renderImportBuddy',
			'args'			=>	array( $state ),
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		$this->insert_next_step( $nextStep );
		
		
		
		
		return true;
	} // End deploy_pull_files().
	
	
	
	/* deploy_getFile()
	 *
	 * Supports chunked retrieval.
	 *
	 * @param	string	$destinationPath	Path to store file in locally. Inlude trailing slash.
	 *
	 */
	public function deploy_getFile( $state, $destinationPath, $sendFile, $sendType, $finalNextStep, $seekTo = 0 ) {
		$TIME_WIGGLE_ROOM = 5; // seconds of wiggle room around sending.
		$numSentThisRound = 0;
		
		/*
		pb_mainwp_backupbuddy::status( 'details', 'Path: `' . $destinationPath . '`.' );
		pb_mainwp_backupbuddy::status( 'details', 'File: `' . $sendFile . '`.' );
		pb_mainwp_backupbuddy::status( 'details', 'Type: `' . $sendType . '`.' );
		*/
		
		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/remote_api.php' );
		
		$timeStart = microtime( true );
		$maxPayload = $state['destinationSettings']['max_payload']; // in MB. convert to bytes.
		$saveFile = $destinationPath . $sendFile;
		$keepSending = true;
		
		while( true == $keepSending ) {
			if ( false === ( $response = mainwp_backupbuddy_remote_api::remoteCall( $state['destination'], 'getFile_' . $sendType, array( 'filename' => $sendFile, 'seekto' => $seekTo, 'maxPayload' => $maxPayload ), $state['minimumExecutionTime'] ) ) ) {
				pb_mainwp_backupbuddy::status( 'error', 'Error #2373273: Unable to initiate file get via remote API.' );
				return false;
			}
			
			if ( ! is_array( $response ) ) {
				pb_mainwp_backupbuddy::status( 'error', 'Error #38937324: Expected return array. Response: `' . htmlentities( $response ) . '`.' );
				return false;
			}
			
			/* $response now contains an array from the server.
			 *	filedata		string		base64 encoded file contents.
			 *	filedatalen		int			length of the filedata param.
			 *	filecrc			string		CRC of the filedata param.
			 *	filedone		1|0			1 if this is the last of the file.
			 *	filesize		int			Total file size in bytes.
			 *	resumepoint		int			ftell resume point to pass to next fseek. 0 if filedone = 1.
			 *	encoded			bool		True if filename needs to be utf8_decoded
			 */
			
			$numSentThisRound++;
			
			if ( ( 0 == $seekTo ) && ( is_file( $saveFile ) ) ) { // First or only part so delete if already exists & is a file (not directory &; not a resumed chunk).
				if ( true !== unlink( $saveFile ) ) {
					$message = 'Error #83832983: Unable to deleting existing local file to overwrite. Check permissions on file/directory `' . $saveFile . '`.';
					pb_mainwp_backupbuddy::status( 'error', $message );
					return false;
				}
			}
			
			// Make sure containing directory exists.
			$saveDir = dirname( $saveFile );
			if ( ! is_dir( $saveDir ) ) {
				if ( true !== pb_mainwp_backupbuddy::$filesystem->mkdir( $saveDir ) ) {
					$message = 'Error #327832: Unable to create directory `' . $saveDir . '`. Check permissions or manually create. Halting to preserve deployment integrity';
					pb_mainwp_backupbuddy::status( 'error', $message );
					return false;
				}
			}
			
			// Open/create file for write/append.
			if ( false === ( $fs = fopen( $saveFile, 'a' ) )) {
				$message = 'Error #43834984: Unable to fopen file `' . $sendFile . '` in directory `' . $$destinationPath . '`.';
				pb_mainwp_backupbuddy::status( 'error', $message );
				return false;
			}
			
			// Seek to specific location. fseeking rather than direct append in case there's some sort of race condition and this gets out of order in some way. Best to be specific.
			if ( 0 != fseek( $fs, $seekTo ) ) {
				@fclose( $fs );
				$message = 'Error #2373792: Unable to fseek file.';
				pb_mainwp_backupbuddy::status( 'error', $message );
				return false;
			}
			
			// Write to file.
			if ( false === ( $bytesWritten = fwrite( $fs, base64_decode( $response['filedata'] ) ) ) ) { // Failed writing.
				@fclose( $fs );
				@unlink( $saveFile );
				$message = 'Error #4873474: Error writing to file `' . $saveFile . '`.';
				pb_mainwp_backupbuddy::status( 'error', $message );
				return false;
			} else { // Success writing.
				@fclose( $fs );
				
				$message = 'Wrote `' . $bytesWritten . '` bytes to `' . $saveFile . '`.';
				pb_mainwp_backupbuddy::status( 'details', $message );
			}
			
			$elapsed = microtime( true ) - $timeStart;
			
			// Handle finishing up or chunking if needed.
			if ( '1' == $response['filedone'] ) { // Transfer finished.
				pb_mainwp_backupbuddy::status( 'deployFileSent', 'File sent.' );
				pb_mainwp_backupbuddy::status( 'details', 'Retrieval of complete file + writing took `' . $elapsed .'` secs.');
				$keepSending = false;
				
				$this->insert_next_step( $finalNextStep );
				return true;
			} else { // More chunks remain.
				pb_mainwp_backupbuddy::status( 'details', 'Retrieval of chunk + writing took `' . $elapsed . '` seconds. Wrote `' . pb_mainwp_backupbuddy::$format->file_size( $bytesWritten ) . '` of max limit of `' . pb_mainwp_backupbuddy::$format->file_size( $maxPayload ) . '` MB per chunk after seeking to `' . $seekTo . '`. Encoded size received: `' . strlen( $response['filedata'] ) . '` bytes.');
				$seekTo = $response['resumepoint']; // Next chunk fseek to this point (got by ftell).
				
				if ( ( ( $elapsed + $TIME_WIGGLE_ROOM ) * ( $numSentThisRound + 1 ) ) >= $state['minimumExecutionTime'] ) { // Could we have time to send another piece based on average send time?
					pb_mainwp_backupbuddy::status( 'details', 'Not enough time to request more data this pass. Chunking.' );
					$keepSending = false;
					
					$nextStep = array(
						'function'		=>	'deploy_getFile',
						'args'			=>	array( $state, $destinationPath, $sendFile, $sendType, $finalNextStep, $seekTo ),
						'start_time'	=>	0,
						'finish_time'	=>	0,
						'attempts'		=>	0,
					);
					
					$this->insert_next_step( $nextStep );
					return true;
				} else {
					pb_mainwp_backupbuddy::status( 'details', 'There appears to be enough time to get more data this pass.' );
					$keepSending = true;
				}
			}
		} // end while( true == $keepSending ).
		
	} // End deploy_getFile().
	
	
	
	/* deploy_pull_renderImportBuddy()
	 *
	 * description
	 *
	 */
	public function deploy_pull_renderImportBuddy( $state ) {
		
//		if ( '' == $state['pullLocalArchiveFile'] ) {
//			pb_mainwp_backupbuddy::status( 'details', 'Skipping rendering of ImportBuddy step because there is no archive file. This is usually due to selecting no database tables to be pulled, therefore no import/migration needed.' );
//			pb_mainwp_backupbuddy::status( 'deployFinished', 'Finished.' );
//			return true;
//		}
//		
//		if ( ! file_exists( $state['pullLocalArchiveFile'] ) ) {
//			pb_mainwp_backupbuddy::status( 'error', 'Error #32783732: Backup file `' . $state['pullLocalArchiveFile'] . '` not found.' );
//			return false;
//		}
//		$backupSerial = mainwp_backupbuddy_core::get_serial_from_file( $state['pullLocalArchiveFile'] );
//		$importbuddyPassword = md5( md5( $state['destination']['key_public'] ) );
//		$siteurl = site_url();
//		
//		$additionalStateInfo = array(
//			'maxExecutionTime' => $state['minimumExecutionTime']
//		);
//		
//		$importFileSerial = mainwp_backupbuddy_core::deploymentImportBuddy( $importbuddyPassword, $state['pullLocalArchiveFile'], $additionalStateInfo );
//		if ( is_array( $importFileSerial ) ) { // Could not generate importbuddy file.
//			return false;
//		}
//		
//		// Store this serial in settings to cleanup any temp db tables in the future with this serial with periodic cleanup.
//		pb_mainwp_backupbuddy::$options['rollback_cleanups'][ $backupSerial ] = time();
//		pb_mainwp_backupbuddy::save();
//		
//		// Create undo file.
//		$undoFile = 'backupbuddy_deploy_undo-' . $backupSerial . '.php';
//		if ( false === copy( pb_mainwp_backupbuddy::plugin_path() . '/classes/_rollback_undo.php', ABSPATH . $undoFile ) ) {
//			$error = 'Error #3289447: Unable to write undo file `' . ABSPATH . $undoFile . '`. Check permissions on directory.' ;
//			pb_mainwp_backupbuddy::status( 'error', $error);
//			return false;
//		}
//		
//		// Start pulling importbuddy log.
//		$importbuddyURLRoot = $siteurl . '/importbuddy-' . $importFileSerial . '.php';
//		$importbuddyLogURL = $importbuddyURLRoot . '?ajax=getDeployLog&v=' . $importbuddyPassword . '&deploy=true'; //$state['destination']['siteurl'] . '/importbuddy/'?ajax=2&v=' . $importbuddyPassword . '&deploy=true; //status-' . $response['importFileSerial'] . '.txt';
//		$importbuddyURL = $importbuddyURLRoot . '?ajax=2&v=' . $importbuddyPassword . '&deploy=true&direction=pull&file=' . basename( $state['pullLocalArchiveFile'] );
//		pb_mainwp_backupbuddy::status( 'details', 'Load importbuddy at `' . $importbuddyURLRoot . '` with verifier `' . $importbuddyPassword . '`.' );
//		pb_mainwp_backupbuddy::status( 'loadImportBuddy', json_encode( array( 'url' => $importbuddyURL, 'logurl' => $importbuddyLogURL ) ) );
//		
//		// Calculate undo URL.
//		$undoDeployURL = $siteurl . '/backupbuddy_deploy_undo-' . $this->_backup['serial'] . '.php';
//		pb_mainwp_backupbuddy::status( 'details', 'To undo deployment of database contents go to the URL: ' . $undoDeployURL );
//		pb_mainwp_backupbuddy::status( 'undoDeployURL', $undoDeployURL );
//		
//		// Pull importbuddy log instead of remote backup log. Nothing else is going to be done on remote server.
//		$this->_backup_options->options['deployment_log'] = $importbuddyLogURL;
//		$this->_backup_options->save();
//		
//		// Next step.
//		pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to run importbuddy steps on remote server.' );
//		$newStep = array(
//			'function'		=>	'deploy_runningImportBuddy',
//			'args'			=>	array( $state ),
//			'start_time'	=>	0,
//			'finish_time'	=>	0,
//			'attempts'		=>	0,
//		);
//		$this->insert_next_step( $newStep );
		
		return true;
	} // End deploy_pull_renderImportBuddy().
	
	
	
	/* deploy_pull_runningBackup()
	 *
	 * Wait for remote backup to complete. Log will be pulled in via AJAX waiting for completion. AJAX will need to tell us once it looks finished so we may proceed.
	 *
	 */
	public function deploy_pull_runningBackup( $state ) {
		
		if ( $this->_backup_options->options['deployment_log'] == get_transient( 'backupbuddy_deployPullBackup_finished' ) ) {
			pb_mainwp_backupbuddy::status( 'details', 'Remote destination backup completed.' );
			
			$this->_backup_options->options['deployment_log'] = ''; // Clear out deployment log so we will not keep hitting the remote server for the backup log, eg while retreiving remote files.
			$this->_backup_options->save();
			
			pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to retrieve files including remote backup archive.' );
			$newStep = array(
				'function'		=>	'deploy_pull_files',
				'args'			=>	array( $state, $pullBackupArchive = $this->_backup_options->options['remote_backup_file'] ),
				'start_time'	=>	0,
				'finish_time'	=>	0,
				'attempts'		=>	0,
			);
			$this->insert_next_step( $newStep );
			
			delete_transient( 'backupbuddy_deployPullBackup_finished' );
			return true;
		}
		
		pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to wait for backup to finish on remote server.' );
		$newStep = array(
			'function'		=>	'deploy_pull_runningBackup',
			'args'			=>	array( $state ),
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		$this->insert_next_step( $newStep );
		
		return true;
	} // End deploy_waitingImportBuddy().
	
	
	
	/* deploy_push_sendContent()
	 *
	 * Used PUSHING.
	 *
	 */
	public function deploy_push_sendContent( $state ) {
		// Next step will be back here after any send file waiting finishes. Keep coming back here until there is nothing more to send.
		$nextStep = array(
			'function'		=>	'deploy_push_sendContent',
			'args'			=>	array(), // MUST populate state before passing off this next step.
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		
		// Count up files to send.
		$mediaFileCount = 0;
		$pluginFileCount = 0;
		$themeFileCount = 0;
		$childThemeFileCount = 0;
		if ( true === $state['sendMedia'] ) {
			$mediaFileCount = count( $state['pushMediaFiles'] );
		}
		if ( count( $state['sendPlugins'] ) > 0 ) {
			$pluginFileCount = count( $state['pushPluginFiles'] );
		}
		if ( true === $state['sendTheme'] ) {
			$themeFileCount = count( $state['pushThemeFiles'] );
		}
		if ( true === $state['sendChildTheme'] ) {
			$childThemeFileCount = count( $state['pushChildThemeFiles'] );
		}
		$filesRemaining = $mediaFileCount + $pluginFileCount + $themeFileCount + $childThemeFileCount;
		pb_mainwp_backupbuddy::status( 'deployFilesRemaining', $filesRemaining );
		pb_mainwp_backupbuddy::status( 'details', 'Files remaining to send: ' . $filesRemaining );
		
		if ( true !== $state['sendMedia'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING push of media files.' );
		} else {
			if ( $mediaFileCount > 0 ) { // Media files remain to send.
				$sendFile = array_pop( $state['pushMediaFiles'] ); // Pop off last item in array. Faster than shift.
				$wp_upload_dir = wp_upload_dir();
				$nextStep['args'] = array( $state );
				return $this->deploy_push_sendFile( $state, $wp_upload_dir['basedir'] . '/' . $sendFile, $sendFile, 'media', $nextStep );
			}
		}
		
		if ( 0 == count( $state['sendPlugins'] ) ) {
			pb_mainwp_backupbuddy::status( 'details', 'No plugin files selected for transfer. Skipping send.' );
		} else {
			if ( $pluginFileCount > 0 ) { // Plugin files remain to send.
				pb_mainwp_backupbuddy::status( 'details', 'Plugins files remaining to send: ' . count( $state['pushPluginFiles'] ) );
				$sendFile = array_pop( $state['pushPluginFiles'] ); // Pop off last item in array. Faster than shift.
				$pluginPath = wp_normalize_path( WP_PLUGIN_DIR ) . '/';
				$nextStep['args'] = array( $state );
				return $this->deploy_push_sendFile( $state, $pluginPath . $sendFile, $sendFile, 'plugin', $nextStep );
			}
		}
		
		if ( true !== $state['sendTheme'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING push of theme files.' );
		} else {
			if ( $themeFileCount > 0 ) { // Plugin files remain to send.
				$sendFile = array_pop( $state['pushThemeFiles'] ); // Pop off last item in array. Faster than shift.
				$themePath = get_template_directory(); // contains trailing slash.
				$nextStep['args'] = array( $state );
				return $this->deploy_push_sendFile( $state, $themePath . $sendFile, $sendFile, 'theme', $nextStep );
			}
		}
		
		if ( true !== $state['sendChildTheme'] ) {
			pb_mainwp_backupbuddy::status( 'details', 'SKIPPING push of child theme files.' );
		} else {
			if ( $childThemeFileCount > 0 ) { // Plugin files remain to send.
				$sendFile = array_pop( $state['pushChildThemeFiles'] ); // Pop off last item in array. Faster than shift.
				$childThemePath = get_stylesheet_directory(); // contains trailing slash.
				$nextStep['args'] = array( $state );
				return $this->deploy_push_sendFile( $state, $childThemePath . $sendFile, $sendFile, 'childTheme', $nextStep );
			}
		}
		
//		// If we made it here then all file sending is finished. Move on to next step.
//		$nextStep = array(
//			'function'		=>	'deploy_push_renderImportBuddy',
//			'args'			=>	array( $state ),
//			'start_time'	=>	0,
//			'finish_time'	=>	0,
//			'attempts'		=>	0,
//		);
//		$this->insert_next_step( $nextStep );
		
		return true;
	} // End deploy_push_sendcontent().
	
	
	
	public function deploy_push_renderImportBuddy( $state ) {
		
//		$timeout = 10;
//		pb_mainwp_backupbuddy::status( 'details', 'Tell remote server to render importbuddy & place our settings file.' );
//		
//		require_once( pb_mainwp_backupbuddy::plugin_path() . '/classes/deploy.php' );
//		$deploy = new mainwp_backupbuddy_deploy( $state['destinationSettings'], $state );
//		
//		if ( false === ( $response = mainwp_backupbuddy_remote_api::remoteCall(
//				$state['destination'],
//				'renderImportBuddy',
//				array(
//					'backupFile' => basename( $this->_backup['archive_file'] ),
//					'max_execution_time' => $state['minimumExecutionTime'],
//				),
//				$state['minimumExecutionTime']
//			) ) ) {
//			pb_mainwp_backupbuddy::status( 'error', 'Error #4448985: Unable to render importbuddy via remote API.' );
//			return false;
//		
//		}
//		pb_mainwp_backupbuddy::status( 'details', 'Render importbuddy result: `' . print_r( $response, true ) . '`.' );
//		
//		// Calculate importbuddy URL.
//		$importbuddyPassword = md5( md5( $state['destination']['key_public'] ) ); // Double md5 like a rainbow.
//		$importbuddyURLRoot = $state['destination']['siteurl'] . '/importbuddy-' . $response['importFileSerial'] . '.php';
//		$importbuddyLogURL = $importbuddyURLRoot . '?ajax=getDeployLog&v=' . $importbuddyPassword . '&deploy=true'; //$state['destination']['siteurl'] . '/importbuddy/'?ajax=2&v=' . $importbuddyPassword . '&deploy=true; //status-' . $response['importFileSerial'] . '.txt';
//		$importbuddyURL = $importbuddyURLRoot . '?ajax=2&v=' . $importbuddyPassword . '&deploy=true&direction=push&file=' . basename( $this->_backup['archive_file'] );
//		pb_mainwp_backupbuddy::status( 'details', 'Load importbuddy at `' . $importbuddyURLRoot . '` with verifier `' . $importbuddyPassword . '`.' );
//		pb_mainwp_backupbuddy::status( 'loadImportBuddy', json_encode( array( 'url' => $importbuddyURL, 'logurl' => $importbuddyLogURL ) ) );
//		
//		// Calculate undo URL.
//		$undoDeployURL = $state['destination']['siteurl'] . '/backupbuddy_deploy_undo-' . $this->_backup['serial'] . '.php';
//		pb_mainwp_backupbuddy::status( 'details', 'To undo deployment of database contents go to the URL: ' . $undoDeployURL );
//		pb_mainwp_backupbuddy::status( 'undoDeployURL', $undoDeployURL );
//		
//		$this->_backup_options->options['deployment_log'] = $importbuddyLogURL;
//		$this->_backup_options->save();
//		
//		pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to run importbuddy steps on remote server.' );
//		$newStep = array(
//			'function'		=>	'deploy_runningImportBuddy',
//			'args'			=>	array( $state ),
//			'start_time'	=>	0,
//			'finish_time'	=>	0,
//			'attempts'		=>	0,
//		);
//		$this->insert_next_step( $newStep );
		
		return true;
		
	} // End deploy_push_renderImportBuddy().
	
	
	
	public function deploy_runningImportBuddy( $state ) {
		
		$maxImportBuddyWaitTime = 60*60*48; // 48 hrs.
		
		// Safety net just in case a loop forms.
		if ( ( time() - $state['startTime'] ) > $maxImportBuddyWaitTime ) {
			pb_mainwp_backupbuddy::status( 'error', 'Error #8349484: Fatal error. ImportBuddy is taking too long to complete. Aborting deployment.' );
			return false;
		}
		
		pb_mainwp_backupbuddy::status( 'details', 'Inserting deploy step to run importbuddy steps on remote server.' );
		$newStep = array(
			'function'		=>	'deploy_runningImportBuddy',
			'args'			=>	array( $state ),
			'start_time'	=>	0,
			'finish_time'	=>	0,
			'attempts'		=>	0,
		);
		$this->insert_next_step( $newStep );
		sleep( 1 ); // Sleep to insure at least a minimum pause between running importbuddy steps.
		return true;
		
	} // End deploy_waitingImportBuddy().
	
	
	/* insert_next_step()
	 *
	 * Inserts a step to run next and saves it to fileoptions.
	 *
	 * $nextStep = step array.
	 */
	private function insert_next_step( $newStep ) {
		array_splice( $this->_backup_options->options['steps'], $this->_currentStepIndex + 1, 0, array( $newStep ) );
		$this->_backup_options->save();
	} // End nextStep().
	
	
	
	/********* BEGIN MULTISITE (Exporting subsite; creates a standalone backup) *********/
	
	
	
	/*	ms_download_extract_wordpress()
	 *	
	 *	Used by Multisite Exporting.
	 *	Downloads and extracts the latest WordPress for making a standalone backup of a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean		True on success, else false.
	 */
	public function ms_download_extract_wordpress() {
		
		return true;
		
	} // End ms_download_wordpress().
	
	
	
	/*	ms_create_wp_config()
	 *	
	 *	Used by Multisite Exporting.
	 *	Creates a standalone wp-config.php file for making a standalone backup from a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean			Currently only returns true.
	 */
	public function ms_create_wp_config() {
		
		pb_mainwp_backupbuddy::status( 'message', 'Creating new wp-config.php file for temporary WordPress installation.' );
		
		global $current_blog;
		$blog_id = absint( $current_blog->blog_id );
		
		//Step 3 - Create new WP-Config File
		$to_file = "<?php\n";
		$to_file .= sprintf( "define( 'DB_NAME', '%s' );\n", '' );
		$to_file .= sprintf( "define( 'DB_USER', '%s' );\n", '' );
		$to_file .= sprintf( "define( 'DB_PASSWORD', '%s' );\n", '' );
		$to_file .= sprintf( "define( 'DB_HOST', '%s' );\n", '' );
		$charset = defined( 'DB_CHARSET' ) ? DB_CHARSET : '';
		$collate = defined( 'DB_COLLATE' ) ? DB_COLLATE : '';
		$to_file .= sprintf( "define( 'DB_CHARSET', '%s' );\n", $charset );
		$to_file .= sprintf( "define( 'DB_COLLATE', '%s' );\n", $collate );
		
		//Attempt to remotely retrieve salts
		$salts = wp_remote_get( 'https://api.wordpress.org/secret-key/1.1/salt/' );
		if ( !is_wp_error( $salts ) ) { // Success.
			$to_file .= wp_remote_retrieve_body( $salts ) . "\n";
		} else { // Failed.
			pb_mainwp_backupbuddy::status( 'warning', 'Error getting salts from WordPress.org for wp-config.php. You may need to manually edit your wp-config on restore. Error: `' . $salts->get_error_message() . '`.' );
		}
		$to_file .= sprintf( "define( 'WPLANG', '%s' );\n", WPLANG );
		$to_file .= sprintf( '$table_prefix = \'%s\';' . "\n", 'bbms' . $blog_id . '_' );
		
		$to_file .= "if ( !defined('ABSPATH') ) { \n\tdefine('ABSPATH', dirname(__FILE__) . '/'); }";
		$to_file .= "/** Sets up WordPress vars and included files. */\n
		require_once(ABSPATH . 'wp-settings.php');";
		$to_file .= "\n?>";
		
		//Create the file, save, and close
		$configFile = $this->_backup['backup_root'] . 'wp-config.php';
		$file_handle = fopen( $configFile, 'w' );
		fwrite( $file_handle, $to_file );
		fclose( $file_handle );
		
		pb_mainwp_backupbuddy::status( 'message', 'Temporary WordPress wp-config.php file created at `' . $configFile . '`.' );
		
		return true;
	} // End ms_create_wp_config().
	
	
	
	/*	ms_copy_plugins()
	 *	
	 *	Used by Multisite Exporting.
	 *	Copies over the selected plugins for inclusion into the backup for creating a standalone backup from a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean			True on success, else false.
	 */
	public function ms_copy_plugins() {
	
		pb_mainwp_backupbuddy::status( 'message', 'Copying selected plugins into temporary WordPress installation.' );
		
		// Step 4 - Copy over plugins.
		// Move over plugins.
		$plugin_items = $this->_backup['export_plugins'];
		
		// Get plugins for site.
		$site_plugins = get_option( 'active_plugins' );
		if ( !empty( $site_plugins ) ) {
			$plugin_items['site'] = $site_plugins;
		}
		
		//Populate $items_to_copy for all plugins to copy over
		if ( is_array( $plugin_items ) ) {
			$items_to_copy = array();
			//Get content directories by using this plugin as a base
			$content_dir = $dropin_plugins_dir = dirname( dirname( dirname( rtrim( plugin_dir_path(__FILE__), '/' ) ) ) );
			$mu_plugins_dir = $content_dir . '/mu-plugins';
			$plugins_dir = $content_dir . '/plugins';
			
			//Get the special plugins (mu, dropins, network activated)
			foreach ( $plugin_items as $type => $plugins ) {
				foreach ( $plugins as $plugin ) {
					if ( $type == 'mu' ) {
						$items_to_copy[ $plugin ] = $mu_plugins_dir . '/' . $plugin;
					} elseif ( $type == 'dropin' ) {
						$items_to_copy[ $plugin ] = $dropin_plugins_dir . '/' . $plugin;
					} elseif ( $type == 'network' || $type == 'site' ) {
						//Determine if we're a folder-based plugin, or a file-based plugin (such as hello.php)
						$plugin_path = dirname( $plugins_dir . '/' . $plugin );
						if ( basename( $plugin_path ) == 'plugins' ) {
							$plugin_path = $plugins_dir . '/' . $plugin;
						}
						$items_to_copy[ basename( $plugin_path ) ] = $plugin_path;		
					}
				} //end foreach $plugins.
			} //end foreach special plugins.
			
			
			//Copy the files over
			$wp_dir = '';
			if ( count( $items_to_copy ) > 0 ) {
				$wp_dir = $this->_backup['backup_root'];
				$wp_plugin_dir = $wp_dir . 'wp-content/plugins/';
				foreach ( $items_to_copy as $file => $original_destination ) {
					if ( file_exists( $original_destination ) && file_exists( $wp_plugin_dir ) ) {
						//$this->copy( $original_destination, $wp_plugin_dir . $file ); 
						$result = pb_mainwp_backupbuddy::$filesystem->recursive_copy( $original_destination, $wp_plugin_dir . $file );
						
						if ( $result === false ) {
							pb_mainwp_backupbuddy::status( 'error', 'Unable to copy plugin from `' . $original_destination . '` to `' . $wp_plugin_dir . $file . '`. Verify permissions.' );
							return false;
						} else {
							pb_mainwp_backupbuddy::status( 'details', 'Copied plugin from `' . $original_destination . '` to `' . $wp_plugin_dir . $file . '`.' );
						}
					}
				}
			}
			
			// Finished
			
			pb_mainwp_backupbuddy::status( 'message', 'Copied selected plugins into temporary WordPress installation.' );
			return true;

		} else {
			// Nothing has technically failed at this point - There just aren't any plugins to copy over.
			
			pb_mainwp_backupbuddy::status( 'message', 'No plugins were selected for backup. Skipping plugin copying.' );
			return true;
		}
		
		return true; // Shouldnt get here.
		
	} // End ms_copy_plugins().
	
	
	
	/*	ms_copy_themes()
	 *	
	 *	Used by Multisite Exporting.
	 *	Copies over the selected themes for inclusion into the backup for creating a standalone backup from a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean			True on success, else false.
	 */
	public function ms_copy_themes() {
	
		
		pb_mainwp_backupbuddy::status( 'message', 'Copying theme(s) into temporary WordPress installation.' );
		
		if ( !function_exists( 'wp_get_theme' ) ) {
			pb_mainwp_backupbuddy::status( 'details', 'wp_get_theme() function not found. Loading `/wp-admin/includes/theme.php`.' );
			require_once( ABSPATH . 'wp-admin/includes/theme.php' );
			pb_mainwp_backupbuddy::status( 'details', 'Loaded `/wp-admin/includes/theme.php`.' );
		}
		
		// Use new wp_get_theme() if available.
		if ( function_exists( 'wp_get_theme' ) ) { // WordPress v3.4 or newer.
			pb_mainwp_backupbuddy::status( 'details', 'wp_get_theme() available. Using it.' );
			$current_theme = wp_get_theme();
		} else { // WordPress pre-v3.4
			pb_mainwp_backupbuddy::status( 'details', 'wp_get_theme() still unavailable (pre WordPress v3.4?). Attempting to use older current_theme_info() fallback.' );
			$current_theme = current_theme_info();
		}
		
				
		//Step 5 - Copy over themes
		$template_dir = $current_theme->template_dir;
		$stylesheet_dir = $current_theme->stylesheet_dir;
		
		pb_mainwp_backupbuddy::status( 'details', 'Got current theme information.' );
		
		//If $template_dir and $stylesheet_dir don't match, that means we have a child theme and need to copy over the parent also
		$items_to_copy = array();
		$items_to_copy[ basename( $template_dir ) ] = $template_dir;
		if ( $template_dir != $stylesheet_dir ) {
			$items_to_copy[ basename( $stylesheet_dir ) ] = $stylesheet_dir;
		}
		
		pb_mainwp_backupbuddy::status( 'details', 'About to begin copying theme files...' );
		
		//Copy the files over
		if ( count( $items_to_copy ) > 0 ) {
			$wp_dir = $this->_backup['backup_root'];
			$wp_theme_dir = $wp_dir . 'wp-content/themes/';
			foreach ( $items_to_copy as $file => $original_destination ) {
				if ( file_exists( $original_destination ) && file_exists( $wp_theme_dir ) ) {
					
					$result = pb_mainwp_backupbuddy::$filesystem->recursive_copy( $original_destination, $wp_theme_dir . $file ); 
					
					if ( $result === false ) {
						pb_mainwp_backupbuddy::status( 'error', 'Unable to copy theme from `' . $original_destination . '` to `' . $wp_theme_dir . $file . '`. Verify permissions.' );
						return false;
					} else {
						pb_mainwp_backupbuddy::status( 'details', 'Copied theme from `' . $original_destination . '` to `' . $wp_theme_dir . $file . '`.' );
					}
				} // end if file exists.
			} // end foreach $items_to_copy.
		} // end if.
		
		pb_mainwp_backupbuddy::status( 'message', 'Copied theme into temporary WordPress installation.' );
		return true;
		
	} // End ms_copy_themes().
	
	
	
	/*	ms_copy_media()
	 *	
	 *	Used by Multisite Exporting.
	 *	Copies over media (wp-content/uploads) for this site for inclusion into the backup for creating a standalone backup from a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean			True on success, else false.
	 */
	public function ms_copy_media() {
		
		pb_mainwp_backupbuddy::status( 'message', 'Copying media into temporary WordPress installation.' );
		
		//Step 6 - Copy over media/upload files
		$upload_dir = wp_upload_dir();
		$original_upload_base_dir = $upload_dir[ 'basedir' ];
		$destination_upload_base_dir = $this->_backup['backup_root'] . 'wp-content/uploads';
		//$result = pb_mainwp_backupbuddy::$filesystem->custom_copy( $original_upload_base_dir, $destination_upload_base_dir, array( 'ignore_files' => array( $this->_backup['serial'] ) ) );
		
		// Grab directory upload contents so we can exclude backupbuddy directories.
		$upload_contents = glob( $original_upload_base_dir . '/*' );
		if ( !is_array( $upload_contents ) ) {
			$upload_contents = array();
		}
				
		foreach( $upload_contents as $upload_content ) {
			if ( strpos( $upload_content, 'backupbuddy_' ) === false ) { // Dont copy over any backupbuddy-prefixed uploads directories.
				$result = pb_mainwp_backupbuddy::$filesystem->recursive_copy( $upload_content, $destination_upload_base_dir . '/' . basename( $upload_content ) );
			}
		}
		
		if ( $result === false ) {
			pb_mainwp_backupbuddy::status( 'error', 'Unable to copy media from `' . $original_upload_base_dir . '` to `' . $destination_upload_base_dir . '`. Verify permissions.' );
			return false;
		} else {
			pb_mainwp_backupbuddy::status( 'details', 'Copied media from `' . $original_upload_base_dir . '` to `' . $destination_upload_base_dir . '`.' );
			return true;
		}
		
	} // End ms_copy_media().
	
	
	
	/*	ms_copy_users_table()
	 *	
	 *  Step 7
	 *	Used by Multisite Exporting.
	 *	Copies over users to a temp table for this site for inclusion into the backup for creating a standalone backup from a subsite.
	 *	Authored by Ron H. Modified by Dustin B.
	 *	
	 *	@return		boolean			Currently only returns true.
	 */
	public function ms_copy_users_table() {
		
		pb_mainwp_backupbuddy::status( 'message', 'Copying temporary users table for users in this blog.' );

		global $wpdb, $current_blog;
		
		$new_user_tablename = $wpdb->prefix . 'users';
		$new_usermeta_tablename = $wpdb->prefix . 'usermeta';
		
		if ( $new_user_tablename == $wpdb->users ) {
			pb_mainwp_backupbuddy::status( 'message', 'Temporary users table would match existing users table. Skipping creation of this temporary users & usermeta tables.' );
			return true;
		}
		
		// Copy over users table to temporary table.
		pb_mainwp_backupbuddy::status( 'message', 'Created new table `' . $new_user_tablename . '` like `' . $wpdb->users . '`.' );
		$wpdb->query( "CREATE TABLE `{$new_user_tablename}` LIKE `{$wpdb->users}`" );
		$wpdb->query( "INSERT `{$new_user_tablename}` SELECT * FROM `{$wpdb->users}`" );
		
		// Copy over usermeta table to temporary table.
		pb_mainwp_backupbuddy::status( 'message', 'Created new table `' . $new_usermeta_tablename . '` like `' . $wpdb->usermeta . '`.' );
		$wpdb->query( "CREATE TABLE `{$new_usermeta_tablename}` LIKE `{$wpdb->usermeta}`" );
		$wpdb->query( "INSERT `{$new_usermeta_tablename}` SELECT * FROM `{$wpdb->usermeta}`" );
		
		// Get list of users associated with this site.
		$users_to_capture = array();
		$user_args = array(
			'blog_id' => $current_blog->blog_id
		);
		$users = get_users( $user_args );
		if ( $users ) {
			foreach ( $users as $user ) {
				array_push( $users_to_capture, $user->ID );
			}
		}
		$users_to_capture = implode( ',', $users_to_capture );
		pb_mainwp_backupbuddy::status( 'details', 'User IDs to capture (' . count( $users_to_capture ) . ' total): ' . print_r( $users_to_capture, true ) );
		
		// Remove users from temporary table that arent associated with this site.
		$wpdb->query( "DELETE from `{$new_user_tablename}` WHERE ID NOT IN( {$users_to_capture} )" );
		$wpdb->query( "DELETE from `{$new_usermeta_tablename}` WHERE user_id NOT IN( {$users_to_capture} )" );
		

		
		pb_mainwp_backupbuddy::status( 'message', 'Copied temporary users table for users in this blog.' );
		return true;
		
	} // End ms_copy_users_table().
	
	public function ms_cleanup() {
		pb_mainwp_backupbuddy::status( 'details', 'Beginning Multisite-export specific cleanup.' );
		
		global $wpdb;
		$new_user_tablename = $wpdb->prefix . 'users';
		$new_usermeta_tablename = $wpdb->prefix . 'usermeta';
		
		if ( ( $new_user_tablename == $wpdb->users ) || ( $new_usermeta_tablename == $wpdb->usermeta ) ) {
			pb_mainwp_backupbuddy::status( 'error', 'Unable to clean up temporary user tables as they match main tables. Skipping to prevent data loss.' );
			return;
		}
		
		pb_mainwp_backupbuddy::status( 'details', 'Dropping temporary table `' . $new_user_tablename . '`.' );
		$wpdb->query( "DROP TABLE `{$new_user_tablename}`" );
		pb_mainwp_backupbuddy::status( 'details', 'Dropping temporary table `' . $new_usermeta_tablename . '`.' );
		$wpdb->query( "DROP TABLE `{$new_usermeta_tablename}`" );
		
		pb_mainwp_backupbuddy::status( 'details', 'Done Multisite-export specific cleanup.' );
	}
	
	/********* END MULTISITE *********/
	
	
	
	
	
	
	
	
	
} // End class.
